/*
 *	cell.h
 *
 *	A cell is some kind of generalized button, which may contain as many
 *	subviews as needed (current maximum is OP_CELL_CTRL). It handles pressing
 *	and/or highlighting.
 *
 *	(C) Copyright 1994-1998, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#ifndef	_OPaC_CELL_H_
#define	_OPaC_CELL_H_

#ifndef _OPaC_OPEXPORT_H_
#include <OPaC/opexport.h>
#endif

OP_INTERFACE (cell.h);

#ifndef _OPaC_CONTROL_H_
#include <OPaC/control.h>
#endif

/****************************************************************/

/*
 *	A cell may have up to 4 border lines (top, left, bottom
 *	and right). 
 */

#define	OP_CELL_BORDER_TOP		0
#define	OP_CELL_BORDER_BOTTOM	1
#define	OP_CELL_BORDER_LEFT		2
#define	OP_CELL_BORDER_RIGHT	3

/*
 *	A cell can contain several controls, which are placed side
 *	by side (horizontally).
 */

#define	OP_CELL_CTRL			4

/****************************************************************/

/*
 *	The controls found in the cell will be either centered,
 *	left aligned or right aligned. The alignment will be
 *	done sequentially in the cell.
 */

OP_TYPEDEF (Card32, OPCellAlignment);

#define	OP_CELL_ALIGN_LEFT		0
#define	OP_CELL_ALIGN_CENTER	1
#define	OP_CELL_ALIGN_RIGHT		2

/****************************************************************/

/*
 *	The cell class contains up to OP_CELL_CTRL controls.
 */

class OP_PUBLIC OPCell : public OPControl
{
	
private:
	Bool				border[4];				//	the 4 edges of the cell can be bordered
	OPControlRef		control[OP_CELL_CTRL];	//	controls which are contained in the cell
	OPCellAlignment		align[OP_CELL_CTRL];	//	alignment for each control
	OPCoord				space[2][OP_CELL_CTRL];	//	space may be added left/right to the controls
	
protected:
	OPCell ();
	virtual ~ OPCell ();
	
	virtual void StoreInstance (OPStorageIn storage);	/* OVERRIDDEN */
	virtual void RestoreInstance (OPStorageIn storage);	/* OVERRIDDEN */
	
	virtual void UpdateGeometry ();
	virtual void ChangeGeometry (OPRectIn);				/* OVERRIDDEN */
	
public:
	virtual void Delete ();								/* OVERRIDDEN */
	virtual Bool EmptyCell ();
	
	virtual Bool SetControl (Count index,
							 OPControlIn control,
							 OPCellAlignment align = OP_CELL_ALIGN_CENTER,
							 OPCoord left_space = 0,
							 OPCoord right_space = 0);
	
	virtual Bool SetControl (OPControlIn control,
							 OPCellAlignment align = OP_CELL_ALIGN_CENTER,
							 OPCoord left_space = 0,
							 OPCoord right_space = 0);
	
	virtual Bool SetAlignment (Count index, OPCellAlignment);
	virtual Bool SetLeftSpace (Count index, OPCoord space);
	virtual Bool SetRightSpace (Count index, OPCoord space);
	
	virtual Bool GetControl (Count index, OPControlOut control) const;
	virtual Bool FindControl (OPControlIn control, CountOut index);
	
	virtual void SetCellBorder (Count which, Bool state);
	virtual Bool GetCellBorder (Count which) const;
	
	//	Inherited methods :
	
	virtual void GetMinSize (OPCoordOut, OPCoordOut) const;				/* OVERRIDDEN */
	virtual void GetMaxSize (OPCoordOut, OPCoordOut) const;				/* OVERRIDDEN */
	
	virtual void Broadcast (OPAtomIn, OPObjectIn);						/* OVERRIDDEN */
	virtual void Draw (OPGrailPortIn, OPRectIn, OPDrawOp = OP_DRAW_ALL);/* OVERRIDDEN */
	
	OP_DEFINE_CLASS (OPCell, OPControl);
};

#endif /* _OPaC_CELL_H_ */

