/*
 *	sd_grail.cxx
 *
 *	This file implements the system dependent interface to the X11 system.
 *
 *	(C) Copyright 1994-1998, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

// Modified by Laurent Bovet 09.02.98

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#include "sd_interface.h"

/****************************************************************/

#include <OPaC/grail.h>
#include <OPaC/lookman.h>
#include <OPaC/new.h>
#include <OPaC/sd/grail.h>

/****************************************************************/

#define	BUFFER_LENGTH	400

/****************************************************************/

OP_MORE_IMPLEMENTATION (OPGrailPort);

/****************************************************************/

/*
 *	Initialise an empty Grail Port.
 */

void
OPaC_GrailPort_Initialise (OPaC_GrPort*& info, OPColorPixel f, OPColorPixel b)
{
	OP_FUNCTION (OPaC_GrailPort_Init, 3);
	OP_ARG (1, OPaC_GrPort*&, info);
	OP_ARG (2, OPColorPixel, f);
	OP_ARG (3, OPColorPixel, b);
	
	XGCValues values;

	info = OP_ALLOC_FROM_ZONE (OPaC_sys_zone, OPaC_GrPort);
	info->xdrawable = 0;
	
	// Create a Graphics Context

	values.foreground = (unsigned long) f;
	values.background = (unsigned long) b;

	info->xgc = XCreateGC(xdpy, RootWindow(xdpy, 0),
												(GCForeground | GCBackground),
												&values);
}


/*
 *	Free a Grail Port.
 */


void
OPaC_GrailPort_Kill (OPaC_GrPort*& info)
{
	OP_FUNCTION (OPaC_GrailPort_Kill, 1);
	OP_ARG (1, OPaC_GrPort*&, info);
		
	XFreeGC(xdpy, info->xgc);
	
	OP_FREE_FROM_ZONE (OPaC_sys_zone, info);
}


/*
 *	Start drawing into the Grail Port. This does not setup
 *	a clipping region...
 */

void
OPaC_GrailPort_BeginDraw (OPaC_GrPort* info, int x1, int y1, int x2, int y2,
						  OPaC_ColorMan* cman)
{
	OP_FUNCTION (OPaC_GrailPort_BeginDraw, 6);
	OP_ARG (1, OPaC_GrPort*, info);
	OP_ARG (2, int, x1);
	OP_ARG (3, int, y1);
	OP_ARG (4, int, x2);
	OP_ARG (5, int, y2);
	OP_ARG (6, OPaC_ColorMan*, cman);
	
	// Nothing must be allocated before drawing...
}


/*
 *	Finish drawing. This may be used to free the resources
 *	allocated by .._BeginDraw
 */

void
OPaC_GrailPort_EndDraw (OPaC_GrPort* info)
{
	OP_FUNCTION (OPaC_GrailPort_EndDraw, 1);
	OP_ARG (1, OPaC_GrPort*, info);
	
	// Nothing must be freed after drawing...
}



/*
 *	Flush the graphical output. This forces the drawing
 *	operations to be executed. This will not display the
 *	 result on the screen, because of the double buffering.
 */

void
OPaC_GrailPort_Flush (OPaC_GrPort* info)
{
	OP_FUNCTION (OPaC_GrailPort_Flush, 1);
	OP_ARG (1, OPaC_GrPort*, info);
	
	XFlush(xdpy);
}



/*
 *	Define a clipping rectangle. This will be called only
 *	between a ..BeginDraw and a ..EndDraw.
 */

void
OPaC_GrailPort_SetClipRect (OPaC_GrPort* info, int x1, int y1, int x2, int y2)
{
	OP_FUNCTION (OPaC_GrailPort_SetClipRect, 5);
	OP_ARG (1, OPaC_GrPort*, info);
	OP_ARG (2, int, x1);
	OP_ARG (3, int, y1);
	OP_ARG (4, int, x2);
	OP_ARG (5, int, y2);
	
	XRectangle xrectangle; 
  
	xrectangle.x      = (short) x1;
	xrectangle.y      = (short) y1;
	xrectangle.width  = (unsigned short)(x2 - x1 + 1);
	xrectangle.height = (unsigned short)(y2 - y1 + 1);

	XSetClipRectangles(xdpy, info->xgc, 0, 0, &xrectangle, 1, Unsorted);
}



/*
 *	Define the foreground color. This will be called only
 *	between a ..BeginDraw and a ..EndDraw.
 */

void
OPaC_GrailPort_SetForeColor (OPaC_GrPort* info, OPColorPixel pixel, OPColorManIn cman)
{
	OP_FUNCTION (OPaC_GrailPort_SetForeColor, 3);
	OP_ARG (1, OPaC_GrPort*, info);
	OP_ARG (2, OPColorPixel, pixel);
	OP_ARG (3, OPColorManIn, cman);
	
	XSetForeground(xdpy, info->xgc, (unsigned long) pixel);
}


/*
 *	Define the background color. This will be called only
 *	between a ..BeginDraw and a ..EndDraw.
 */

void
OPaC_GrailPort_SetBackColor (OPaC_GrPort* info, OPColorPixel pixel, OPColorManIn cman)
{
	OP_FUNCTION (OPaC_GrailPort_SetBackColor, 3);
	OP_ARG (1, OPaC_GrPort*, info);
	OP_ARG (2, OPColorPixel, pixel);
	OP_ARG (3, OPColorManIn, cman);
	
	XSetBackground(xdpy, info->xgc, (unsigned long) pixel);
}


/*
 *	Define the line style.  This will be called only
 *	between a ..BeginDraw and a ..EndDraw.
 */

void
OPaC_GrailPort_SetLineStyle (OPaC_GrPort* info, OPColorPixel pixel, OPLineStyle style, OPColorManIn cman)
{
	OP_FUNCTION (OPaC_GrailPort_SetLineStyle, 4);
	OP_ARG (1, OPaC_GrPort*, info);
	OP_ARG (2, OPColorPixel, pixel);
	OP_ARG (3, OPLineStyle, style);
	OP_ARG (4, OPColorManIn, cman);
	
	const int pen_style[] = { LineSolid, LineOnOffDash, LineOnOffDash };

	const char dash_list[2][2] = { {1, 1},   // Dotted :  - - - - -
															 {4, 4} }; // Dashed :    ----    ----    ----

	OP_ASSERT(info->xgc);

	XSetDashes(xdpy, info->xgc, 0, 
						 style == OP_LINE_DOT ? dash_list[0] : dash_list[1],
						 2);                    // Two values specify the dash_list
						 
	XSetForeground(xdpy, info->xgc, (unsigned long) pixel);
	XSetLineAttributes(xdpy, info->xgc,
										 0,                // Line width -> thin hair line
										 pen_style[style],
										 CapButt,          // How to end lines (not usefull since width=1)
										 JoinMiter);      // How connected lines join (same remark)
}

/*
 *	Define the current font.  This will be called only
 *	between a ..BeginDraw and a ..EndDraw.
 */

void
OPaC_GrailPort_SetFont (OPaC_GrPort* info, OPaC_GrFont* font_info)
{
	OP_FUNCTION (OPaC_GrailPort_SetFont, 2);
	OP_ARG (1, OPaC_GrPort*, info);
	OP_ARG (2, OPaC_GrFont*, font_info);

	if(font_info && font_info->xfont)
		XSetFont(xdpy, info->xgc, font_info->xfont);
	else {
		Font xfont = XLoadFont(xdpy, "Fixed");  // Default font
		XSetFont(xdpy, info->xgc, xfont);
	}
}

/****************************************************************
 *																*
 *																*
 *			GRAIL PORT CLASS : TARGET SPECIFIC METHODS			*
 *																*
 *																*
 ****************************************************************/

/*
 *	Draw just one pixel.
 */

void
OPGrailPort::DrawDot (OPCoord x, OPCoord y)
{
	OP_METHOD (DrawDot, 2);
	OP_ARG (1, OPCoord, x);
	OP_ARG (2, OPCoord, y);
	
	XDrawPoint(xdpy, info->xdrawable, info->xgc, (int)x, (int)y );
}

void
OPGrailPort::DrawDot (OPCoord x, OPCoord y, OPColorPixel color)
{
	OP_METHOD (DrawDot, 3);
	OP_ARG (1, OPCoord, x);
	OP_ARG (2, OPCoord, y);
	OP_ARG (3, OPColorPixel, color);
	
	XGCValues values;

	XGetGCValues(xdpy, info->xgc, GCForeground, &values); // Save previous color
	XSetForeground(xdpy, info->xgc, (unsigned long) color);
	XDrawPoint(xdpy, info->xdrawable, info->xgc, (int)x, (int)y );
	XChangeGC(xdpy, info->xgc, GCForeground, &values); // Restore previous color
}

/*
 *	Draw a line (segment), using one of the predefined dash
 *	styles. The segment is always hair thin... The last dot
 *	will be drawn.
 *
 *	A line length of 0 won't draw anything !
 */

void
OPGrailPort::DrawLineDX (OPCoord x, OPCoord y, OPCoord dx)
{
	OP_METHOD (DrawLineDX, 3);
	OP_ARG (1, OPCoord, x);
	OP_ARG (2, OPCoord, y);
	OP_ARG (3, OPCoord, dx);
	OP_ASSERT (this->is_drawing == TRUE);
	
	if (dx != 0) 
		XDrawLine(xdpy, info->xdrawable, info->xgc, (int)x, (int)y, (int)(x+dx-1), (int)y);
}

void
OPGrailPort::DrawLineDY (OPCoord x, OPCoord y, OPCoord dy)
{
	OP_METHOD (DrawLineDY, 3);
	OP_ARG (1, OPCoord, x);
	OP_ARG (2, OPCoord, y);
	OP_ARG (3, OPCoord, dy);
	OP_ASSERT (this->is_drawing == TRUE);
	
	if (dy != 0) 
		XDrawLine(xdpy, info->xdrawable, info->xgc, (int)x, (int)y, (int)x, (int)(y+dy-1));
}



/*
 *	Draw a line (segment), using one of the predefined dash
 *	styles. The segment is always hair thin... The last dot
 *	will be drawn.
 *
 *	A line with both extremities at the same location will
 *	draw just a single dot.
 */

void
OPGrailPort::DrawLineToX (OPCoord x1, OPCoord y1, OPCoord x2)
{
	OP_METHOD (DrawLineToX, 3);
	OP_ARG (1, OPCoord, x1);
	OP_ARG (2, OPCoord, y1);
	OP_ARG (3, OPCoord, x2);
	OP_ASSERT (this->is_drawing == TRUE);
	
	if (x2 >= x1) {
		XDrawLine(xdpy, info->xdrawable, info->xgc, (int)x1, (int)y1, (int)x2-1, (int)y1);
	} else {
		XDrawLine(xdpy, info->xdrawable, info->xgc, (int)x2+1, (int)y1, (int)x1, (int)y1);		
	}
}

void
OPGrailPort::DrawLineToY (OPCoord x1, OPCoord y1, OPCoord y2)
{
	OP_METHOD (DrawLineToY, 3);
	OP_ARG (1, OPCoord, x1);
	OP_ARG (2, OPCoord, y1);
	OP_ARG (3, OPCoord, y2);
	OP_ASSERT (this->is_drawing == TRUE);
	
	if (y2 >= y1) {
		XDrawLine(xdpy, info->xdrawable, info->xgc, (int)x1, (int)y1, (int)x1, (int)y2-1);		
	} else {
		XDrawLine(xdpy, info->xdrawable, info->xgc, (int)x1, (int)y2+1, (int)x1, (int)y1);		
	}
}


/*
 *	Draw rectangular box. The bottom left corner will be
 *	drawn at [x2;y2]. This might not be the case on every
 *	system, so be careful to implent this the OPaC way.
 */

void
OPGrailPort::DrawRect (OPRectIn box, OPRelCoord margin)
{
	OP_METHOD (DrawRect, 2);
	OP_ARG (1, OPRectIn, box);
	OP_ARG (2, OPRelCoord, margin);
	OP_ASSERT (this->is_drawing == TRUE);
	
	if ( ((OPRelCoord)(box.DX ()) < (2*margin))
	  || ((OPRelCoord)(box.DY ()) < (2*margin)) ) {
		return;
	}
	
	OPCoord x1 = box.X1 () + margin;
	OPCoord y1 = box.Y1 () + margin;
	OPCoord x2 = box.X2 () - margin;
	OPCoord y2 = box.Y2 () - margin;
	
	XDrawRectangle(xdpy, info->xdrawable, info->xgc, 
								 (int)x1, (int)y1,
								 (unsigned int)(x2-x1),  (unsigned int)(y2-y1));
}

void
OPGrailPort::DrawRect (OPCoord x1, OPCoord y1, OPCoord x2, OPCoord y2)
{
	OP_METHOD (DrawRect, 4);
	OP_ARG (1, OPCoord, x1);
	OP_ARG (2, OPCoord, y1);
	OP_ARG (3, OPCoord, x2);
	OP_ARG (4, OPCoord, y2);
	OP_ASSERT (this->is_drawing == TRUE);

	XDrawRectangle(xdpy, info->xdrawable, info->xgc,
								 (int)x1, (int)y1, 
								 (unsigned int)(x2-x1),  (unsigned int)(y2-y1));
}



/*
 *	Fill a rectangular box. The surface drawn is exactly
 *	the same as with DrawRect. Beware, this is usually not
 *	the case on other systems !!!
 *
 *	BE CAREFUL TO IMPLEMENT THIS METHOD THE RIGHT WAY !
 */

void
OPGrailPort::FillRect (OPRectIn box, OPRelCoord margin)
{
	OP_METHOD (FillRect, 2);
	OP_ARG (1, OPRectIn, box);
	OP_ARG (2, OPRelCoord, margin);
	OP_ASSERT (this->is_drawing == TRUE);
	
	if ( ((OPRelCoord)(box.DX ()) < (2*margin))
	  || ((OPRelCoord)(box.DY ()) < (2*margin)) ) {
		return;
	}
	
	OPCoord x1 = box.X1 () + margin;
	OPCoord y1 = box.Y1 () + margin;
	OPCoord x2 = box.X2 () - margin;
	OPCoord y2 = box.Y2 () - margin;

	XFillRectangle(xdpy, info->xdrawable, info->xgc, 
								 (int)x1, (int)y1,
								 (unsigned int)(x2-x1+1), (unsigned int)(y2-y1+1));	
}

void
OPGrailPort::FillRect (OPCoord x1, OPCoord y1, OPCoord x2, OPCoord y2)
{
	OP_METHOD (FillRect, 4);
	OP_ARG (1, OPCoord, x1);
	OP_ARG (2, OPCoord, y1);
	OP_ARG (3, OPCoord, x2);
	OP_ARG (4, OPCoord, y2);
	OP_ASSERT (this->is_drawing == TRUE);
	
	XFillRectangle(xdpy, info->xdrawable, info->xgc,
								 (int)x1, (int)y1,
								 (unsigned int)(x2-x1+1), (unsigned int)(y2-y1+1));
}


/*
 *	Draw a line segment. The segment is always hair thin and
 *	is not dashed... The last dot will be drawn.
 */

void
OPGrailPort::DrawHairLine (OPCoord x, OPCoord y, OPCoord dx, OPCoord dy)
{
	OP_METHOD (DrawHairLine, 4);
	OP_ARG (1, OPCoord, x);
	OP_ARG (2, OPCoord, y);
	OP_ARG (3, OPCoord, dx);
	OP_ARG (4, OPCoord, dy);
	
	this->DrawHairLineTo (x, y, x+dx, y+dy);
}


void
OPGrailPort::DrawHairLineTo (OPCoord x1, OPCoord y1, OPCoord x2, OPCoord y2)
{
	OP_METHOD (DrawHairLineTo, 4);
	OP_ARG (1, OPCoord, x1);
	OP_ARG (2, OPCoord, y1);
	OP_ARG (3, OPCoord, x2);
	OP_ARG (4, OPCoord, y2);

	XGCValues values;

	XGetGCValues(xdpy, info->xgc, 
							 (GCLineWidth | GCLineStyle), // Save previous line attributes
							 &values); 
	XSetLineAttributes(xdpy, info->xgc, 0, LineSolid, CapButt, JoinMiter);
	XDrawLine(xdpy, info->xdrawable, info->xgc, (int)x1, (int)y1, (int)x2, (int)y2);
	XChangeGC(xdpy, info->xgc, GCForeground, &values); // Restore previous color
}



/*
 *	Draw a line segment with anti-aliasing. The segment is always
 *	hair thin and is not dashed... The last dot will be drawn.
 */

void
OPGrailPort::DrawAALine (OPCoord x, OPCoord y, OPCoord dx, OPCoord dy)
{
	OP_METHOD (DrawAALine, 4);
	OP_ARG (1, OPCoord, x);
	OP_ARG (2, OPCoord, y);
	OP_ARG (3, OPCoord, dx);
	OP_ARG (4, OPCoord, dy);
	
	this->DrawAALineTo (x, y, x+dx, y+dy);
}

void
OPGrailPort::DrawAALineTo (OPCoord x1, OPCoord y1, OPCoord x2, OPCoord y2)
{
	OP_METHOD (DrawAALineTo, 4);
	OP_ARG (1, OPCoord, x1);
	OP_ARG (2, OPCoord, y1);
	OP_ARG (3, OPCoord, x2);
	OP_ARG (4, OPCoord, y2);

	// TODO: Implement in a little bit more right way...

	XDrawLine(xdpy, info->xdrawable, info->xgc, (int)x1, (int)y1, (int)x2, (int)y2);
}

/*
 *	Draw a piece of text, which will be extracted from a
 *	string starting at character "start" and up to "length"
 *	characters. The way to position the text is specified
 *	by "pos" and "dx" is used if aligned positioning has
 *	been requested.
 */

void
OPGrailPort::ShowText (OPCoord x, OPCoord y, OPStringIn text,
					   Count start, Count length,
					   Card8 pos,
					   OPCoord dx)
{
	OP_METHOD (ShowText, 7);
	OP_ARG (1, OPCoord, x);
	OP_ARG (2, OPCoord, y);
	OP_ARG (3, OPStringIn, text);
	OP_ARG (4, Count, start);
	OP_ARG (5, Count, length);
	OP_ARG (6, Card8, pos);
	OP_ARG (7, OPCoord, dx);
	OP_ASSERT (this->is_drawing == TRUE);
	
	int xx = (int)x, xy = (int)y;

	Count len = text->CountCharacters ();

	if (start >= len) {
		
		//	Do not draw anything, since the start of the text in the string is
		//	past the end of the string.
		
		return;
	}
	
	if ( (len < (start + length))
			 || (length == 0) ) {
		
		//	User is asking for a too long piece of text, longer than it really
		//	can be.
		
		length = len - start;
	}
	
	LatinChar  buffer_array[BUFFER_LENGTH];
	LatinChar* buffer_alloc = 0;
	LatinChar* buffer = 0;
	
	if (length < BUFFER_LENGTH) {
		
		//	Use the buffer from the stack.
		
		buffer = buffer_array;
		
	} else {
		
		//	Use the buffer from the temporary memory.
		
		buffer = OP_ALLOC_ARRAY_FROM_ZONE (OPaC_temp_zone, LatinChar, length);
		buffer_alloc = buffer;
	}
	
	text->ExtractStringCopy (buffer, start, length);
	
	XCharStruct properties; // overall characteristics of the string
	XGCValues values;
	int direction, ascent, descent;
	
	XGetGCValues(xdpy, info->xgc, GCFont, &values); // Get the current font ID

	XQueryTextExtents( xdpy, values.font, buffer, length,
										 &direction, &ascent, &descent,
										 &properties); // Get the characteristics of the string
	
	int hi = (pos & OP_TEXT_HPOS_MASK);
	int vi = (pos & OP_TEXT_VPOS_MASK);
	
	if (hi == OP_TEXT_HPOS_ALIGN) {
		
		//	Text should be aligned on both margins.
		//	NIY
		
		hi = OP_TEXT_HPOS_CENTER;
	}
	
	switch(vi) {
	case OP_TEXT_VPOS_TOP    : xy += ascent; break;
	case OP_TEXT_VPOS_CENTER : xy += ascent - (ascent + descent) /2; break;
	case OP_TEXT_VPOS_BOTTOM : xy -= descent; break;
	}

	switch(hi) {
	case OP_TEXT_HPOS_CENTER : xx -= properties.width / 2; break;
	case OP_TEXT_HPOS_RIGHT  : xx -= properties.width; break;
	case OP_TEXT_HPOS_ALIGN  : xx -= properties.width / 2; break;  
		// Yet same as ..CENTER. Should be different.
	}

	XDrawString(xdpy, info->xdrawable, info->xgc,
							(int)xx, (int)xy,
							(char*) buffer,
							(int) length);

	if (buffer_alloc) {
		
		//	Free the memory if we need to.
		
		OP_FREE_FROM_ZONE (OPaC_temp_zone, buffer_alloc);
	}
}



/*
 *	Draw a piece of text in the specified rectangle. This
 *	uses the above ShowText method
 */

void
OPGrailPort::ShowText (OPRectIn box, OPStringIn text,
					   Count start, Count length, Card8 pos, OPRelCoord margin)
{
	OP_METHOD (ShowText, 6);
	OP_ARG (1, OPRectIn, box);
	OP_ARG (2, OPStringIn, text);
	OP_ARG (3, Count, start);
	OP_ARG (4, Count, length);
	OP_ARG (5, Card8, pos);
	OP_ARG (6, OPRelCoord, margin);
	OP_ASSERT (this->is_drawing == TRUE);
	

	if ( ((OPRelCoord)(box.DX ()) < (2*margin))
	  || ((OPRelCoord)(box.DY ()) < (2*margin)) ) {
		return;
	}
	
	OPCoord x1 = box.X1 () + margin;
	OPCoord y1 = box.Y1 () + margin;
	OPCoord x2 = box.X2 () - margin;
	OPCoord y2 = box.Y2 () - margin;
	
	this->ShowText (x1, y1, x2, y2, text, start, length, pos);
}

void
OPGrailPort::ShowText (OPCoord x1, OPCoord y1, OPCoord x2, OPCoord y2,
					   OPStringIn text,
					   Count start, Count length,
					   Card8 pos)
{
	OP_METHOD (ShowText, 8);
	OP_ARG (1, OPCoord, x1);
	OP_ARG (2, OPCoord, y1);
	OP_ARG (3, OPCoord, x2);
	OP_ARG (4, OPCoord, y2);
	OP_ARG (5, OPStringIn, text);
	OP_ARG (6, Count, start);
	OP_ARG (7, Count, length);
	OP_ARG (8, Card8, pos);
	OP_ASSERT (this->is_drawing == TRUE);
	
	if ( (x2 <= x1) || (y2 <= y1) ) {
		return;
	}
	
	OPCoord x = x1;
	OPCoord y = y1;
	
	switch (pos & OP_TEXT_VPOS_MASK) {
		
		case OP_TEXT_VPOS_TOP:
			y = y1;
			break;
		
		case OP_TEXT_VPOS_CENTER:
			y = (y1 + y2 + 1) / 2;
			break;
			
		case OP_TEXT_VPOS_BOTTOM:
			y = y2;
			break;
	}
	
	switch (pos & OP_TEXT_HPOS_MASK) {
		
		case OP_TEXT_HPOS_LEFT:
			x = x1;
			break;
		
		case OP_TEXT_HPOS_CENTER:
			x = (x1 + x2 + 1) / 2;
			break;
		
		case OP_TEXT_HPOS_ALIGN:
			x = x1;
			break;
		
		case OP_TEXT_HPOS_RIGHT:
			x = x2;
			break;
	}
	
	this->ShowText (x, y, text, start, length, pos, x2-x1+1);
}



/*
 *	Scroll a rectangular area of the Grail Port.
 */

void
OPGrailPort::ScrollRect (OPCoord x1, OPCoord y1, OPCoord x2, OPCoord y2,
						 OPRelCoord move_x, OPRelCoord move_y)
{
	OP_METHOD (ScrollRect, 6);
	OP_ARG (1, OPCoord, x1);
	OP_ARG (2, OPCoord, y1);
	OP_ARG (3, OPCoord, x2);
	OP_ARG (4, OPCoord, y2);
	OP_ARG (5, OPRelCoord, move_x);
	OP_ARG (6, OPRelCoord, move_y);
	OP_ASSERT (this->is_drawing == TRUE);

	// NYI
}

void
OPGrailPort::ScrollRect (OPRectIn box, OPRelCoord move_x, OPRelCoord move_y,
						 OPRelCoord margin)
{
	OP_METHOD (ScrollRect, 4);
	OP_ARG (1, OPRectIn, box);
	OP_ARG (2, OPRelCoord, move_x);
	OP_ARG (3, OPRelCoord, move_y);
	OP_ARG (4, OPRelCoord, margin);
	OP_ASSERT (this->is_drawing == TRUE);
	
	if ( ((OPRelCoord)(box.DX ()) < (2*margin))
	  || ((OPRelCoord)(box.DY ()) < (2*margin)) ) {
		return;
	}
	
	OPCoord x1 = box.X1 () + margin;
	OPCoord y1 = box.Y1 () + margin;
	OPCoord x2 = box.X2 () - margin;
	OPCoord y2 = box.Y2 () - margin;
	
	this->ScrollRect (x1, y1, x2, y2, move_x, move_y);
}



/*
 *	Copy a rectangle from the source port to the current
 *	port, into the specified box, from the source position
 *	given by its top/left corner.
 *
 *	NB: This will copy the pixels, including the one of
 *		the bottom right corner. This is still the same
 *		policy as with the rectangle.
 */

void
OPGrailPort::CopyRect (OPGrailPortIn source,
					   OPCoord source_x, OPCoord source_y,
					   OPCoord dest_x1, OPCoord dest_y1,
					   OPCoord dest_x2, OPCoord dest_y2)
{
	OP_METHOD (CopyRect, 7);
	OP_ARG (1, OPGrailPortIn, source);
	OP_ARG (2, OPCoord, source_x);
	OP_ARG (3, OPCoord, source_y);
	OP_ARG (4, OPCoord, dest_x1);
	OP_ARG (5, OPCoord, dest_y1);
	OP_ARG (6, OPCoord, dest_x2);
	OP_ARG (7, OPCoord, dest_y2);
	OP_ASSERT (this->is_drawing == TRUE);

	unsigned int width = (unsigned int)(dest_x2-dest_x1+1);
	unsigned int height = (unsigned int)(dest_y2-dest_y1+1);

	if(this->use_transparency) {

		GC gc = XCreateGC(xdpy, info->xdrawable, 0, 0);

    // Temporary pixmaps

		Pixmap colormask = XCreatePixmap(xdpy, DefaultRootWindow(xdpy),
																		 width,
																		 height,
																		 DefaultDepth(xdpy, DefaultScreen(xdpy)));

		Pixmap clipmask = XCreatePixmap(xdpy, DefaultRootWindow(xdpy),
																		width,
																		height,
																		1); // This is a bitmap
		
		Pixmap tempdest = XCreatePixmap(xdpy, DefaultRootWindow(xdpy),
																		(unsigned int) dest_x2+1,
																		(unsigned int) dest_y2+1,
																		DefaultDepth(xdpy, DefaultScreen(xdpy)));

		// Fill the colormask with the transparent color

		XSetForeground(xdpy, gc, this->transparent_color);
		XFillRectangle(xdpy, colormask, gc, 0, 0,
									 width,
									 height);

		// Copy the source to the colormask with an equiv function
		// It means that each transparent pixel will have all its bits set to 1
		// Other pixels will have other values...
	
		XGCValues colormask_values;
		colormask_values.function = GXequiv;
		GC colormask_gc = 
			XCreateGC(xdpy, colormask, GCFunction, &colormask_values);

		XCopyArea(xdpy, source->info->xdrawable, colormask, colormask_gc, 
							(int)source_x, (int)source_y,
							width, height,
							0, 0);

		// Merge each plane of the colormask into the clipmask with an "and 
		// inverted" function. There will be 1's for transparent pixels and 0
		// for others.

		XGCValues clipmask_values;
		clipmask_values.function = GXset;
		GC clipmask_gc = XCreateGC(xdpy, clipmask, GCFunction, &clipmask_values);

		XFillRectangle(xdpy, clipmask, clipmask_gc, 0, 0,
									 width,
									 height);

		clipmask_values.function = GXandInverted;
		XChangeGC(xdpy, clipmask_gc, GCFunction, &clipmask_values);

		for(int i=1; i < (1 << DefaultDepth(xdpy, DefaultScreen(xdpy))); i*=2) {
			XCopyPlane(xdpy,
								 colormask,
								 clipmask,
								 clipmask_gc,
								 0,
								 0,
								 width,
								 height,
								 0,
								 0,
								 (unsigned long) i);
		}

		// Inverts the clipmask
		clipmask_values.function = GXcopyInverted;
		XChangeGC(xdpy, clipmask_gc, GCFunction, &clipmask_values);
		
		XCopyArea(xdpy, clipmask, clipmask, clipmask_gc,
							0, 0,
							(unsigned int)(dest_x2-dest_x1+1),
							(unsigned int)(dest_y2-dest_y1+1),			
							0, 0);
		
		XGCValues dest_values;
		dest_values.clip_mask = clipmask;
		dest_values.clip_x_origin = dest_x1;
		dest_values.clip_y_origin = dest_y1;
		
		XChangeGC(xdpy, gc, 
							GCClipMask | GCClipXOrigin | GCClipYOrigin,
							&dest_values);
		
		// Copy the old part of the destination to the tempdest

		XCopyArea(xdpy, info->xdrawable, tempdest, info->xgc, 
							(int)dest_x1,
							(int)dest_y1,
							(unsigned int)(dest_x2-dest_x1+1), 
							(unsigned int)(dest_y2-dest_y1+1),
							(int)(dest_x1),
							(int)(dest_y1));

		// Copy the rectangle to tempdest using the mask
		
		XCopyArea(xdpy, source->info->xdrawable, tempdest, gc, 
							(int)source_x,
							(int)source_y,
							(unsigned int)(dest_x2-dest_x1+1), 
							(unsigned int)(dest_y2-dest_y1+1),
							(int)(dest_x1),
							(int)(dest_y1));

		// Finally copy the part of tempdest to the window

		XCopyArea(xdpy, tempdest, info->xdrawable, info->xgc, 
							(int)dest_x1,
							(int)dest_y1,
							(unsigned int)(dest_x2-dest_x1+1), 
							(unsigned int)(dest_y2-dest_y1+1),
							(int)(dest_x1),
							(int)(dest_y1));
		
		XFlush(xdpy);
		
		// Free allocated resources
		
		XFreeGC(xdpy, gc);
		XFreeGC(xdpy, colormask_gc);
		XFreeGC(xdpy, clipmask_gc);
		
		XFreePixmap(xdpy, colormask);
		XFreePixmap(xdpy, clipmask);
		XFreePixmap(xdpy, tempdest);
	}
	else	
		XCopyArea(xdpy, source->info->xdrawable, info->xdrawable, info->xgc, 
							(int)source_x, (int)source_y,
							width, height,
							(int)(dest_x1), (int)(dest_y1));
}

void
OPGrailPort::CopyRect (OPGrailPortIn source,
					   OPCoord source_x, OPCoord source_y,
					   OPRectIn dest_box, OPRelCoord dest_margin)
{
	OP_METHOD (CopyRect, 5);
	OP_ARG (1, OPGrailPortIn, source);
	OP_ARG (2, OPCoord, source_x);
	OP_ARG (3, OPCoord, source_y);
	OP_ARG (4, OPRectIn, dest_box);
	OP_ARG (5, OPRelCoord, dest_margin);
	OP_ASSERT (this->is_drawing == TRUE);
	
	if ( ((OPRelCoord)(dest_box.DX ()) < (2*dest_margin))
	  || ((OPRelCoord)(dest_box.DY ()) < (2*dest_margin)) ) {
		return;
	}
	
	OPCoord x1 = dest_box.X1 () + dest_margin;
	OPCoord y1 = dest_box.Y1 () + dest_margin;
	OPCoord x2 = dest_box.X2 () - dest_margin;
	OPCoord y2 = dest_box.Y2 () - dest_margin;
	
	this->CopyRect (source, source_x, source_y, x1, y1, x2, y2);
}


/*
 *	Set a line of pixels with the specified pixel values. This
 *	is not very fast, but provides an easy way to build some
 *	pixel maps yourself.
 */

void
OPGrailPort::SetColorPixels (OPCoord x, OPCoord y, OPCoord dx,
							 const OPColorPixel* pixels)
{
	OP_METHOD (SetColorPixels, 4);
	OP_ARG (1, OPCoord, x);
	OP_ARG (2, OPCoord, y);
	OP_ARG (3, OPCoord, dx);
	OP_ARG (4, const OPColorPixel*, pixels);
	OP_ASSERT (this->is_drawing == TRUE);
	
	Count delta = (Count)(dx);
	XGCValues values;

	XGetGCValues(xdpy, info->xgc, GCForeground, &values); // Save previous color

	while (delta--) {
		XSetForeground(xdpy, info->xgc, (unsigned long)(*pixels++));
		XDrawPoint(xdpy, info->xdrawable, info->xgc, (int)x, (int)y );
		x += 1;
	}

	XChangeGC(xdpy, info->xgc, GCForeground, &values); // Restore previous color
}


/*
 *	Set a line of pixels with the specified gray values. This
 *	is not very fast, but provides an easy way to build some
 *	pixel maps yourself.
 */

void
OPGrailPort::SetGrayPixels (OPCoord x, OPCoord y, OPCoord dx,
							const Card8* intensity)
{
	OP_METHOD (SetGrayPixels, 4);
	OP_ARG (1, OPCoord, x);
	OP_ARG (2, OPCoord, y);
	OP_ARG (3, OPCoord, dx);
	OP_ARG (4, cont Card8*, intensity);
	OP_ASSERT (this->is_drawing == TRUE);
	
	this->SetRGBPixels (x, y, dx, intensity, intensity, intensity);
}



/*
 *	Set a line of pixels with the specified RGB values. This
 *	is not very fast, but provides an easy way to build some
 *	pixel maps yourself. The slowness is proportional to the number
 *  of different colors in the line. Not applicable for viewing complex
 *  images...
 */

void
OPGrailPort::SetRGBPixels (OPCoord x, OPCoord y, OPCoord dx,
						   const Card8* red,
						   const Card8* green,
						   const Card8* blue)
{
	OP_METHOD (SetRGBPixels, 6);
	OP_ARG (1, OPCoord, x);
	OP_ARG (2, OPCoord, y);
	OP_ARG (3, OPCoord, dx);
	OP_ARG (4, const Card8*, red);
	OP_ARG (5, const Card8*, green);
	OP_ARG (6, const Card8*, blue);
	OP_ASSERT (this->is_drawing == TRUE);
	
	Count delta = (Count)(dx);
	Count max_len = (Count)dx;

	XColor visited[max_len];  // Remind the colors already allocated
	int nb_visited = 0;

	XColor previous_color;
	XColor current_color;

  XImage* ximage = XGetImage(xdpy, info->xdrawable, 
                             (int)x, (int)y, 
                             (unsigned int)dx, 1,
                             AllPlanes,
                             ZPixmap);

	for(Count j=0; j<delta; j++) {
		current_color.red = (red[j] << 8); // X colors component are 16 bits values
		current_color.green = (green[j] << 8);
		current_color.blue = (blue[j] << 8);

		// If this is the first pixel
		if(j == 0) {
			// Allocate a color for it
			previous_color = current_color;
			visited[0] = current_color;
			XAllocColor(xdpy, XDefaultColormap(xdpy, 0), &current_color);
			visited[nb_visited].pixel = current_color.pixel;
			nb_visited++;			
		}
		else {
			// If the color is the same that for the pixel before, do nothing
			if( !( current_color.red == previous_color.red &&
						 current_color.green == previous_color.green &&
						 current_color.blue == previous_color.blue) ) { 
				
				// Else, look for it in the visited colors table
				Bool found = FALSE;
				for( int i=0; i<nb_visited; i++ )
					if( current_color.red == visited[i].red &&
							current_color.green == visited[i].green &&
							current_color.blue == visited[i].blue ) {
						// If found, we have our color
						previous_color = current_color;
						current_color.pixel = visited[i].pixel;
						found = TRUE;
						break;
					}

				// If not found, allocate it
				if( !found ) {							
					previous_color = current_color;
					visited[nb_visited] = current_color;
					XAllocColor(xdpy, XDefaultColormap(xdpy, 0), &current_color);
					visited[nb_visited].pixel = current_color.pixel;
					nb_visited++;
				}
			}
		}
		
    XPutPixel(ximage, (int)j, (int)0, current_color.pixel);
	}
	
  // Finally send the image to the server
  
  XPutImage(xdpy,
            info->xdrawable, info->xgc, 
            ximage, 
            (int)0, (int)0,
            (int)x, (int)y,
            (int)dx, 1);

  XDestroyImage(ximage);
}


/*
 *	Set a block of pixels with the specified monochrome values.
 *	The pixels set to "1" will appear as black on screen.
 */

void
OPGrailPort::SetBitmap (OPCoord x, OPCoord y, OPCoord dx, OPCoord dy, const Card8* bits)
{
	OP_METHOD (SetBitmap, 5);
	OP_ARG (1, OPCoord, x);
	OP_ARG (2, OPCoord, y);
	OP_ARG (3, OPCoord, dx);
	OP_ARG (3, OPCoord, dy);
	OP_ARG (5, cont Card8*, bits);
	OP_ASSERT (this->is_drawing == TRUE);
	
	//	TODO: set a blok of pixels with the monochrome values defined by "bits"
	//	When "1", use the foreground color, when "0", do not draw anything.
	//	The bits are aligned on an 8-bit boundary (each scanline is byte
	//	aligned).
	printf("Went through unimplemented function SetBitmap\n");
}
