/*
 *	archive.cxx
 *
 *	Archive for saving and restoring data contained in OPaC
 *	objects. This is one very used kind of storage class.
 *
 *	(C) Copyright 1994-1997, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#include <OPaC/archive.h>
#include <OPaC/stream.h>
#include <OPaC/ieeemath.h>
#include <OPaC/new.h>

#include <string.h>

/****************************************************************/

OP_IMPLEMENT_CLASS (OPArchive);

/****************************************************************/

/*
 *	Create a new archive object. Set up an empty state.
 */

OPArchive::OPArchive ()
{
	OP_METHOD (OPArchive, 0);
	
	this->SetStorageMode (STORE_DEEP);
	
	this->self   = 0;			//	still no reference to self...
	this->input  = 0;
	this->output = 0;
}



/*
 *	Destroy an archive. This will flush the remaining data
 *	in the output stream (if any).
 */

OPArchive::~OPArchive ()
{
	OP_METHOD (~OPArchive, 0);
	
	if (this->output.IsValid ()) {
		this->output->Sync ();
	}
}



/*
 *	We cannot possible store an archiving object.
 */

void
OPArchive::StoreInstance (OPStorageIn storage)
{
	OP_METHOD (StoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	OP_ABORT ("Trying to store an archive !");
}



/*
 *	We cannot possible restore an archiving object.
 */

void
OPArchive::RestoreInstance (OPStorageIn storage)
{
	OP_METHOD (RestoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	OP_ABORT ("Trying to restore an archive !");
}



/*
 *	Free any pending links to other objects. This will for
 *	instance remove the link to self, but not the streams,
 *	since they have to be flushed by the destructor.
 */

void
OPArchive::Delete ()
{
	OP_METHOD (Delete, 0);
	
	super::Delete ();
	
	this->input  = 0;
	this->output = 0;
	this->self   = 0;
}



/*
 *	Define the input stream to use for restoring data. It
 *	should be a bufferized byte stream (eg. a file).
 */

void
OPArchive::SetInput (OPStreamIn stream)
{
	OP_METHOD (SetInput, 1);
	OP_ARG (1, OPStreamIn, stream);
	
	LatinChar header[32];
	
	this->input = stream;
	this->self  = (this->input.IsValid ()) ? this : 0;
	this->RestoreData (header, 32);
	
	OP_ASSERT (strncmp (header, "OPaC-BIN", 8) == 0);
	OP_ASSERT ((header[8] == 0) && (header[9] == 0) && (header[10] == 0) && (header[11] == 0));
}



/*
 *	Define the output stream to use for saving data. It
 *	should be a bufferized byte stream (eg. a file).
 */

void
OPArchive::SetOutput (OPStreamIn stream)
{
	OP_METHOD (SetOutput, 1);
	OP_ARG (1, OPStreamIn, stream);
	
	LatinChar header[32];
	bzero (header, 32);
	strncpy (header+0,  "OPaC-BIN", 8);
	strncpy (header+16, "binary archive", 14);
	
	this->output = stream;
	this->self   = (this->output.IsValid ()) ? this : 0;
	
	this->StoreData (header, 32);
}



/*
 *	Reset the input/output streams : this will remove any
 *	remaining stream.
 */

void
OPArchive::Reset ()
{
	OP_METHOD (Reset, 0);
	
	this->input  = 0;
	this->output = 0;
	this->self   = 0;
}



/*
 *	Store any data as raw bytes. Beware of endianness if
 *	you use this method.
 */

void
OPArchive::StoreData (const void* data, Card32 size)
{
	OP_METHOD (SaveData, 2);
	OP_ARG (1, const void*, data);
	OP_ARG (2, Card32, size);
	OP_ASSERT (this->output.IsValid ());
	this->output->Write (data, size);
}



/*
 *	Restore any data as raw bytes. Beware of endianness if
 *	you use this method.
 */

void
OPArchive::RestoreData (void* data, Card32 size)
{
	OP_METHOD (RestoreData, 2);
	OP_ARG (1, void*, data);
	OP_ARG (2, Card32, size);
	OP_ASSERT (this->input.IsValid ());
	this->input->Read (data, size);
}

