/*
 *	classman.cxx
 *
 *	This file implements the OPaC Class Manager library.
 *
 *	(C) Copyright 1994-1998, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#include <OPaC/except.h>
#include <OPaC/object.h>
#include <OPaC/new.h>

/****************************************************************/

Count					OPClassMan::class_num  = 0;
Bool					OPClassMan::is_init_ok = FALSE;
OPClassMan::ClassDef*	OPClassMan::list_head  = 0;
OPClassMan::ClassDef*	OPClassMan::list_tail  = 0;
OPRuntime*				OPClassMan::real_head  = 0;
OPRuntime*				OPClassMan::real_tail  = 0;

/****************************************************************/

/*
 *	Initialise the class manager. This will look for every
 *	recorded class and build the corresponding run-time
 *	information.
 */

void
OPClassMan::Initialise ()
{
	OP_MEMBER (Initialise, 0);
	OP_ASSERT (OPClassMan::is_init_ok == FALSE);
	OP_ASSERT (OPaC_sys_zone);
	
	OPaC_RuntimeInfo* info = 0;
	
	OPRuntime* runtime = 0;
	OPRuntime* head    = 0;
	OPRuntime* tail    = 0;
	
	//	First of all, create the run-time instances. We cannot initialise them
	//	at this point, since the super-classes may not be ready yet.
	
	info = OPaC_RuntimeInfo::last;
	
	while (info) {
		
		//	We cannot instanciate OPRuntime using the OP_NEW or NEW macros, since
		//	the class manager isn't ready yet and since we need to initialise the
		//	"mem_zone" member before calling the constructor !
		
		Card32 size = sizeof (OPRuntime);
		void*  data = OPaC_sys_zone->Alloc (size, "OPRuntime", __LINE__, __FILE__, OP_MTYPE_OBJECT);
		
		((OPObject*)(data))->mem_zone = OPaC_sys_zone;
		
		//	Call the constructor using the pre-allocated buffer. This is OK.
		
		runtime = NEW_USING (data) OPRuntime ();
		runtime->next = 0;
		runtime->mem_zone = OPaC_sys_zone;
		
		(tail ? tail->next : head) = runtime;
		(tail                    ) = runtime;
		
		info = info->prev;
	}
	
	//	Now that every run-time class information has been created, we may
	//	initialised them one step further.
	
	runtime = head;
	info    = OPaC_RuntimeInfo::last;
	
	while (runtime) {
		runtime->Initialise (info);
		runtime = runtime->next;
		info    = info->prev;
	}
	
	//	We may now setup the super-classes, since they are guaranteed to have
	//	been initialised properly too.
	
	runtime = head;
	
	while (runtime) {
		runtime->LinkSuper ();
		runtime = runtime->next;
	}
	
	//	Collect information about the outlets and the targets...
	
	runtime = head;
	
	while (runtime) {
		runtime->CollectTargets ();
		runtime = runtime->next;
	}
	
	//	Compact the collected information into something more "consitent"
	
	runtime = head;
	
	while (runtime) {
		runtime->CompileWhileInit ();
		runtime = runtime->next;
	}
	
	//	We may now insert the run-time class descriptions into the Class Man's
	//	own list.
	
	runtime = head;
	
	while (runtime) {
		
		//	Allocate a run-time class definition slot. It will be linked into
		//	the class manager :
		
		ClassDef* def = NEW_FROM_ZONE (OPaC_sys_zone) ClassDef;
		
		def->prev    = list_tail;
		def->next    = 0;
		def->runtime = runtime->self;
		def->name    = runtime->ReturnClassName ();
		
		(list_tail ? list_tail->next : list_head) = def;
		(list_tail                              ) = def;
		
		class_num++;
		
		runtime = runtime->next;
	}
	
	//	The class manager is now ready to be used.
	
	real_head  = head;
	real_tail  = tail;
	
	OPClassMan::is_init_ok = TRUE;
	
	//	Last step of the initialisation: give once more the opportunity to the
	//	runtime classes to compile more internal information.
	
	runtime = head;
	
	while (runtime) {
		runtime->CompileAfterInit ();
		runtime = runtime->next;
	}
}



/*
 *  When leaving an OPaC application, don't forget to kill
 *	the class manager !
 */

void
OPClassMan::Kill ()
{
	OP_MEMBER (Kill, 0);
	
	OPRuntime* runtime = OPClassMan::real_head;
	
	while (runtime) {
		
		OPRuntime* next = runtime->next;
		
		runtime->Delete ();
		
		runtime = next;
	}
	
	ClassDef* cdef = OPClassMan::list_head;
	
	while (cdef) {
		ClassDef* next = cdef->next;
		DELETE_FROM_ZONE (OPaC_sys_zone, cdef);
		cdef = next;
	}
	
	//	This should free everything. But this would be really complicated, so
	//	we don't do anything special...
	
	class_num  = 0;
	
	list_head  = 0;
	list_tail  = 0;
	
	real_head  = 0;
	real_tail  = 0;
	
	OPClassMan::is_init_ok = FALSE;
	
	//	NIY
}



/*
 *	Return the nth run-time class description. The order
 *	in which the informations are returned is not defined.
 */

Bool
OPClassMan::GetClass (Count pos, OPRuntimeOut runtime)
{
	OP_MEMBER (GetClass, 2);
	OP_ARG (1, Count, pos);
	OP_ARG (2, OPRuntimeOut, runtime);
	
	if (pos >= class_num) {
		runtime = 0;
		return FALSE;
	}
	
	OPRuntime* iter = real_head;
	
	while (pos--) {
		iter = iter->next;
	}
	
	runtime = iter;
	return TRUE;
}



/*
 *	Return the class description corresponding to the
 *	specified name.
 */

Bool
OPClassMan::GetClass (OPAtomIn name, OPRuntimeOut runtime)
{
	OP_MEMBER (GetClass, 2);
	OP_ARG (1, OPAtomIn, name);
	OP_ARG (2, OPRuntimeOut, runtime);
	
	Count     idx  = 0;
	Count     pos  = class_num;
	ClassDef* iter = list_head;
	
	while (pos--) {
		
		if (iter->name == name) {
			
			//	Found the matching class description in the list. Now remove the
			//	element and insert it in the head of the list, so that next time,
			//	it will be found faster.
			
			if (idx > 3) {
				
				//	All this need not be done if the class description is not
				//	really far from the list head...
				
				(iter->next ? iter->next->prev : list_tail) = iter->prev;
				(iter->prev ? iter->prev->next : list_head) = iter->next;
				
				iter->next = list_head;
				iter->prev = 0;
				
				(iter->next ? iter->next->prev : list_tail) = iter;
				(iter->prev ? iter->prev->next : list_head) = iter;
			}
			
			runtime = iter->runtime;
			return TRUE;
		}
		
		iter = iter->next;
		idx += 1;
	}
	
	runtime = 0;
	return FALSE;
}



/*
 *	Return the class description corresponding to the
 *	specified name. First transform the name into an
 *	atom...
 */

Bool
OPClassMan::GetClass (OPStringIn name, OPRuntimeOut runtime)
{
	OP_MEMBER (GetClass, 2);
	OP_ARG (1, OPStringIn, name);
	OP_ARG (2, OPRuntimeOut, runtime);
	
	OPAtomRef atomic_name   = name;
	Bool      runtime_found = GetClass (atomic_name, runtime);
	
	return runtime_found;
}



/*
 *	Return the class description corresponding to the
 *	specified name. First transform the name into an
 *	atom...
 */

Bool
OPClassMan::GetClass (const LatinChar* name, OPRuntimeOut runtime)
{
	OP_MEMBER (GetClass, 2);
	OP_ARG (1, const LatinChar*, name);
	OP_ARG (2, OPRuntimeOut, runtime);
	
	OPAtomRef atomic_name   = name;
	Bool      runtime_found = GetClass (atomic_name, runtime);
	
	return runtime_found;
}



/*
 *	Allocate an instance of the specified class. This will
 *	not be very fast. You had better use the version where
 *	the run-time class is specified directly.
 */

OPObject*
OPClassMan::New (OPAtomIn name, OPMemZone* zone,
				 Count line, const LatinChar* file)
{
	OP_MEMBER (New, 4);
	OP_ARG (1, OPAtomIn, name);
	OP_ARG (2, OPMemZone*, zone);
	OP_ARG (3, Count, line);
	OP_ARG (4, const LatinChar*, file);
	
	OPRuntimeRef runtime;
	Bool ok = GetClass (name, runtime);
	
	if (ok == FALSE) {
		THROW (OPExcObject (OPExcObject::NOT_FOUND));
	}
	
	if (zone == 0) {
		zone = OPaC_app_zone;
	}
	
	Card32 size = runtime->ReturnInstanceSize ();
	void*  data = zone->Alloc (size, runtime->ReturnClassName (), line, file, OP_MTYPE_OBJECT);
	
	ok = runtime->ConstructObject (data, size);
	OP_ASSERT (ok);
	OP_ASSERT (((OPObject*)(data))->mem_zone == zone);
	
	return (OPObject*)(data);
}



/*
 *	
 */

OPObject*
OPClassMan::New (OPStringIn name, OPMemZone* zone,
				 Count line, const LatinChar* file)
{
	OP_MEMBER (New, 4);
	OP_ARG (1, OPStringIn, name);
	OP_ARG (2, OPMemZone*, zone);
	OP_ARG (3, Count, line);
	OP_ARG (4, const LatinChar*, file);
	
	OPAtomRef atom   = name;
	OPObject* object = New (atom, zone, line, file);
	
	return object;
}



/*
 *	
 */

OPObject*
OPClassMan::New (OPRuntimeIn name, OPMemZone* zone,
				 Count line, const LatinChar* file)
{
	OP_MEMBER (New, 4);
	OP_ARG (1, OPRuntimeIn, name);
	OP_ARG (2, OPMemZone*, zone);
	OP_ARG (3, Count, line);
	OP_ARG (4, const LatinChar*, file);
	
	Bool ok = FALSE;
	
	if (zone == 0) {
		zone = OPaC_app_zone;
	}
	
	Card32 size = name->ReturnInstanceSize ();
	void*  data = zone->Alloc (size, name->ReturnClassName (), line, file, OP_MTYPE_OBJECT);
	
	ok = name->ConstructObject (data, size);
	
	OP_ASSERT (ok);
	
	if (name->IsKindOf (OP_RUNTIME (OPAtom))) {
		return (OPObject*)(data);
	}
	
	OP_ASSERT (((OPObject*)(data))->mem_zone == zone);
	
	return (OPObject*)(data);
}



/*
 *	
 */

OPObject*
OPClassMan::New (const LatinChar* name, OPMemZone* zone,
				 Count line, const LatinChar* file)
{
	OP_MEMBER (New, 4);
	OP_ARG (1, const LatinChar*, name);
	OP_ARG (2, OPMemZone*, zone);
	OP_ARG (3, Count, line);
	OP_ARG (4, const LatinChar*, file);
	
	OPAtomRef atom   = name;
	OPObject* object = New (atom, zone, line, file);
	
	return object;
}


