/*
 *	control.cxx
 *
 *	Control class definition. The control class is a widely
 *	used base class for lots of widgets.
 *
 *	(C) Copyright 1994-1998, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#include <OPaC/app.h>
#include <OPaC/control.h>
#include <OPaC/event.h>
#include <OPaC/grail.h>
#include <OPaC/lookman.h>
#include <OPaC/new.h>
#include <OPaC/notmacro.h>
#include <OPaC/pubmacro.h>
#include <OPaC/window.h>

/****************************************************************/

OP_IMPLEMENT_CLASS (OPControl);

/****************************************************************/

OP_NOTIFY_EVENT (OPControl, OnClick);
OP_NOTIFY_EVENT (OPControl, OnContClick);

OP_NOTIFY_CHANGE (OPControl, ActiveHilite, OnClick);

/****************************************************************/

static const OPaC_BitsetDataDef local_shape_info[] = {
	{ 0xFFFFFFFF,			"Shape style",			OP_BITSET_DATA_MASK },
	{ OP_CTRL_RELIEF,		"3D look",				OP_BITSET_DATA_BIT },
	{ OP_CTRL_ABOVE,		"above (if 3D look)",	OP_BITSET_DATA_BIT },
	{ OP_CTRL_EXTRUDE,		"extruded",				OP_BITSET_DATA_BIT },
	{ OP_CTRL_NOFRAME,		"no frame",				OP_BITSET_DATA_BIT },
	{ OP_CTRL_AUTOHILITE,	"automatic hilite",		OP_BITSET_DATA_BIT },
	{ OP_CTRL_AUTOACTHILITE,"auto-active hilite",	OP_BITSET_DATA_BIT },
	{ OP_CTRL_AUTOHOT,		"show when hot",		OP_BITSET_DATA_BIT },
	{ OP_CTRL_AUTOACTIVE,	"show when active",		OP_BITSET_DATA_BIT },
	{ OP_CTRL_AUTODRAG,		"show if drag source",	OP_BITSET_DATA_BIT },
	{ OP_CTRL_AUTODROP,		"show if drop target",	OP_BITSET_DATA_BIT },
	{ OP_CTRL_NOTMAINRESP,	"not main responder",	OP_BITSET_DATA_BIT },
	{ OP_CTRL_NOTACTENTER,	"no activate on enter",	OP_BITSET_DATA_BIT },
	{ OP_CTRL_BROADHILITE,	"broadcast hilite",		OP_BITSET_DATA_BIT }
};

static const OPaC_BitsetDataDef local_sensitive_buttons_info[] = {
	{ 0xFFFFFFFF,			"Mouse buttons",		OP_BITSET_DATA_MASK },
	{ OP_EV_BUTTON_LEFT,	"left button",			OP_BITSET_DATA_BIT },
	{ OP_EV_BUTTON_RIGHT,	"right button",			OP_BITSET_DATA_BIT },
	{ OP_EV_BUTTON_MIDDLE,	"middle button",		OP_BITSET_DATA_BIT },
	{ OP_EV_BUTTON_EXTRA,	"extra button",			OP_BITSET_DATA_BIT }
};

/****************************************************************/

OP_PUBLISH_MODIFUNC (OPControl, "HiliteMode", ReturnHiliteMode, SetHiliteMode, "Hilite mode", Bool, Bool);
OP_PUBLISH_MODIFUNC (OPControl, "PressMode",  ReturnPressMode,  SetPressMode,  "Press mode", Bool, Bool);
OP_PUBLISH_MODIFUNC (OPControl, "ReportMode", ReturnReportMode, SetReportMode, "Continuous report", Bool, Bool);
OP_PUBLISH_MODIFUNC (OPControl, "ActiveHilite", ReturnActiveHilite, SetActiveHilite, "Active hilite", Bool, Bool);
OP_PUBLISH_MODIFUNC (OPControl, "ShapeStyle", ReturnShapeStyle, SetShapeStyle, "Shape style", Card32, Card32);
OP_PUBLISH_MODIFUNC (OPControl, "SensitiveButtons", ReturnSensitiveButtons, SetSensitiveButtons, "Sensitive buttons", Card32, Card32);

OP_PUBLISH_BITSET_INFO (OPControl, "ShapeStyle", local_shape_info);
OP_PUBLISH_BITSET_INFO (OPControl, "SensitiveButtons", local_sensitive_buttons_info);

/****************************************************************/

#define	CONTINUOUS_MS	(OPLookMan::Current ()->ReturnAutoRepeatTime ())

/****************************************************************/

/*
 *	The control class provides a default behavior for many
 *	widget classes. It can be parametrized by setting the
 *	boolean fields at creation.
 */

OPControl::OPControl ()
{
	OP_METHOD (OPControl, 0);
	
	this->is_app_ticking = FALSE;
	
	this->shape = 0;
	
	this->is_being_hilited    = FALSE;
	this->is_being_pressed    = FALSE;
	this->is_being_clicked    = FALSE;
	this->is_actively_hilited = FALSE;
	
	this->do_hilite_on_action       = FALSE;
	this->do_press_on_action        = FALSE;
	this->do_cont_reports_on_action = FALSE;
	
	this->state |= OP_VIEW_OK_DETECT;
	this->state |= OP_VIEW_OK_MAIN;
	
	this->key_alt_known     = FALSE;
	
	this->key_alt.type      = OP_KEY_NONE;
	this->key_alt.value     = 0;
	this->key_alt.code      = 0;
	this->key_alt.modifiers = 0;
	this->sensitive_buttons = OP_EV_BUTTON_LEFT;
	
	this->min_dx = 0;
	this->min_dy = 0;
}



/*
 *	Remove the responder from the application tick list if
 *	it is still subscribed, when destroying.
 */

OPControl::~OPControl ()
{
	OP_METHOD (~OPControl, 0);
	
	if (this->is_app_ticking) {
		this->DisableTick ();
	}
}



/*
 *	Save the control to the storage.
 */

void
OPControl::StoreInstance (OPStorageIn storage)
{
	OP_METHOD (StoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	super::StoreInstance (storage);
	
	storage << this->is_being_hilited;
	storage << this->is_being_pressed;
	storage << this->is_being_clicked;
	storage << this->is_actively_hilited;
	
	storage << this->do_hilite_on_action;
	storage << this->do_press_on_action;
	storage << this->do_cont_reports_on_action;
	
	storage << this->key_alt.type;
	storage << this->key_alt.value;
	storage << this->key_alt.code;
	storage << this->key_alt.modifiers;
	
	storage << this->min_dx << this->min_dy;
	storage << this->shape;
	
	storage << OPStorageSync::mark;
}



/*
 *	Restore the control from the storage.
 */

void
OPControl::RestoreInstance (OPStorageIn storage)
{
	OP_METHOD (RestoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	if (this->key_alt_known) {
		this->ClearKeyAlt ();
	}

	super::RestoreInstance (storage);
	
	storage >> this->is_being_hilited;
	storage >> this->is_being_pressed;
	storage >> this->is_being_clicked;
	storage >> this->is_actively_hilited;

	storage >> this->do_hilite_on_action;
	storage >> this->do_press_on_action;
	storage >> this->do_cont_reports_on_action;
	
	storage >> this->key_alt.type;
	storage >> this->key_alt.value;
	storage >> this->key_alt.code;
	storage >> this->key_alt.modifiers;
	
	storage >> this->min_dx >> this->min_dy;
	storage >> this->shape;
	
	storage >> OPStorageSync::mark;
}

void
OPControl::EnableTick ()
{
	OP_METHOD (EnableTick, 0);
	
	if (this->is_app_ticking) {
		return;
	}
	
	if (this->app.IsValid ()) {
		this->app->InsertTickResponder (this, CONTINUOUS_MS);
		this->is_app_ticking = TRUE;
	}
}

void
OPControl::DisableTick ()
{
	if (this->is_app_ticking) {
		if (this->app.IsValid ()) {
			this->app->RemoveTickResponder (this);
		}
		this->is_app_ticking = FALSE;
	}
}


/*
 *	Prepare the control for destruction.
 */

void
OPControl::Delete ()
{
	OP_METHOD (Delete, 0);
	
	this->DisableTick ();
	
	if (this->key_alt_known) {
		this->ClearKeyAlt ();
	}
	
	super::Delete ();
}



/*
 *	Define the window this control resides in.
 */

Bool
OPControl::SetWindow (OPWindowIn window)
{
	OP_METHOD (SetWindow, 1);
	OP_ARG (1, OPWindowIn, window);
	
	//	This might change the application (not very probable).
	
	OPAppRef no_app;
	OPAppRef new_app = window.IsValid () ? window->ReturnApp () : no_app;
	
	if (this->is_app_ticking) {
		this->DisableTick ();
		this->app = new_app;
		this->EnableTick ();
	} else {
		this->app = new_app;
	}
	
	OPKey key_alt = this->key_alt;
	
	//	Maybe the window won't change, in that case, there is no reason to
	//	worry about keyboard alternative changes.

	if (this->window == window) {
		return super::SetWindow (window);
	}
	
	//	Well, if there was an active keyboard alternative and a known window,
	//	then we have to un-hook ourself from the window's "key.alt" list.
	
	if ( (this->key_alt_known)
	  && (this->window.IsValid ()) ) {
		this->window->NotifyRemoveKeyAlt (this);
		this->key_alt_known = FALSE;
	}
	
	Bool window_ok = super::SetWindow (window);
	
	//	In case we got a valid window and we have a keyboard alternative, just
	//	tell the window about it.
	
	if ( (key_alt.type != OP_KEY_NONE)
	  && (window_ok)
	  && (this->window.IsValid ()) ) {
		this->SetKeyAlt (key_alt);
	}

	return window_ok;
}

	
/*
 *	When the user clicks or performs a similar action on a
 *	control, it will highlight or animate...
 *
 *	This method will NOT redraw the control.
 */

void
OPControl::ActionOn ()
{
	OP_METHOD (ActionOn, 0);
	
	if (this->state & OP_VIEW_IS_ENABLED) {
		
		this->is_being_clicked = TRUE;
		
		if ((this->state & OP_VIEW_IS_ACTIVE) == 0) {
			this->window->DisableMainResponder ();
		}
		
		if (this->do_hilite_on_action) {
			this->SetHiliteState (TRUE, (this->shape & OP_CTRL_BROADHILITE) ? TRUE : FALSE);
		}
		
		if (this->do_press_on_action) {
			this->SetPressState (TRUE, FALSE);
		}
		
		if (this->do_cont_reports_on_action) {
			this->EnableTick ();
		}
	}
}



/*
 *	When the user click ends, the control has to return to
 *	a neutral state back again.
 *
 *	This method will NOT redraw the control.
 */

void
OPControl::ActionOff ()
{
	OP_METHOD (ActionOff, 0);
	
	if (this->state & OP_VIEW_IS_ENABLED) {
		
		this->window->EnableMainResponder ();
		this->is_being_clicked = FALSE;
		
		if (this->do_hilite_on_action) {
			this->SetHiliteState (FALSE, (this->shape & OP_CTRL_BROADHILITE) ? TRUE : FALSE);
		}
		
		if (this->do_press_on_action) {
			this->SetPressState (FALSE, FALSE);
		}
		
		if (this->do_cont_reports_on_action) {
			this->DisableTick ();
		}
		
		this->Display ();
	}
}



/*
 *	The "Action" method gets called after the "ActionOff"
 *	method if the user validated the action.
 */

void
OPControl::Action (Bool clicked)
{
	OP_METHOD (Action, 1);
	OP_ARG (1, Bool, clicked);
	
	if ((this->shape & OP_CTRL_NOTMAINRESP) == 0) {
		this->window->SetMainResponder (OPControlRef (this), clicked);
	}
	
	this->Report (notify_OnClick, clicked ? OP_REPORT_NEW : OP_REPORT_AUTO);
}



/*
 *	Hilite the control (and its children if "broadcast"
 *	has been set). This does not redraw anything. No
 *	action will be taken if the control's state needs no
 *	change.
 */

void
OPControl::SetHiliteState (Bool ok, Bool broadcast)
{
	OP_METHOD (SetHiliteState, 2);
	OP_ARG (1, Bool, ok);
	OP_ARG (2, Bool, broadcast);
	
	if (this->is_being_hilited == ok) {
		return;
	}
	
	this->is_being_hilited = ok;
	
	if (broadcast == FALSE) {
		return;
	}
	
	OPViewRef child;
	this->GetFirstChild (child);
	
	if (child.IsValid ()) {
		
		do {
			
			if (child->IsKindOf (OP_RUNTIME (OPControl))) {
				
				//	Tell the control to change its hiliting state... Simple views
				//	would not know what to do !
				
				OPControlRef control = child;
				control->SetHiliteState (ok, TRUE);
				
			}
			
		} while (child->GetNextSibling (child));
	}
}



/*
 *	Press the control (and its children if "broadcast"
 *	has been set). This does not redraw anything. No
 *	action will be taken if the control's state needs no
 *	change.
 */

void
OPControl::SetPressState (Bool ok, Bool broadcast)
{
	OP_METHOD (SetPressState, 2);
	OP_ARG (1, Bool, ok);
	OP_ARG (2, Bool, broadcast);
	
	if (this->is_being_pressed == ok) {
		return;
	}
	
	this->is_being_pressed = ok;
	
	if (broadcast == FALSE) {
		return;
	}
	
	OPViewRef child;
	this->GetFirstChild (child);
	
	if (child.IsValid ()) {
		
		do {
			
			if (child->IsKindOf (OP_RUNTIME (OPControl))) {
				
				//	Tell the control to change its "press" state... Simple views
				//	would not know what to do !
				
				OPControlRef control = child;
				control->SetPressState (ok, TRUE);
				
			}
			
		} while (child->GetNextSibling (child));
	}
}


/*
 *	Define the control to be hilited actively. This does
 *	does not depend on the automatic hilite state...
 */

Bool
OPControl::SetActiveHilite (Bool ok, Bool broadcast)
{
	OP_METHOD (SetActiveHilite, 2);
	OP_ARG (1, Bool, ok);
	OP_ARG (2, Bool, broadcast);
	
	if (this->is_actively_hilited == ok) {
		return TRUE;
	}
	
	this->is_actively_hilited = ok;
	
	if (broadcast == FALSE) {
		return TRUE;
	}
	
	OPViewRef child;
	this->GetFirstChild (child);
	
	if (child.IsValid ()) {
		
		do {
			
			if (child->IsKindOf (OP_RUNTIME (OPControl))) {
				
				//	Tell the control to change its "press" state... Simple views
				//	would not know what to do !
				
				OPControlRef control = child;
				control->SetActiveHilite (ok, TRUE);
				
			}
			
		} while (child->GetNextSibling (child));
	}
	
	return TRUE;
}

Bool
OPControl::ReturnActiveHilite () const
{
	OP_METHOD (ReturnActiveHilite, 0);
	return this->is_actively_hilited;
}


/*
 *	Enable the continuous event generation as long as the
 *	the control is being pressed.
 */

Bool
OPControl::SetReportMode (Bool is_continuous)
{
	OP_METHOD (SetReportMode, 1);
	OP_ARG (1, Bool, is_continuous);
	
	if (this->do_cont_reports_on_action == is_continuous) {
		return TRUE;
	}
	
	this->do_cont_reports_on_action = is_continuous;
	
	if (is_continuous == FALSE) {
		this->DisableTick ();
	}
	
	if ( (is_continuous == TRUE)
	  && (this->is_being_clicked) ) {
		this->EnableTick ();
	}
	
	return TRUE;
}


/*
 *	Set/get the hilite and press mode.
 */

Bool
OPControl::SetHiliteMode (Bool yes)
{
	OP_METHOD (SetHiliteMode, 1);
	OP_ARG (1, Bool, yes);
	this->do_hilite_on_action = yes;
	return TRUE;
}

Bool
OPControl::SetPressMode (Bool yes)
{
	OP_METHOD (SetPressMode, 1);
	OP_ARG (1, Bool, yes);
	this->do_press_on_action = yes;
	return TRUE;
}

Bool
OPControl::ReturnHiliteMode () const
{
	OP_METHOD (ReturnHiliteMode, 0);
	return this->do_hilite_on_action;
}

Bool
OPControl::ReturnPressMode () const
{
	OP_METHOD (ReturnPressMode, 0);
	return this->do_press_on_action;
}

Bool
OPControl::SetSensitiveButtons (Card32 buttons)
{
	OP_METHOD (SetSensitiveButtons, 1);
	OP_ARG (1, Card32, buttons);
	this->sensitive_buttons = buttons;
	return TRUE;
}

Card32
OPControl::ReturnSensitiveButtons () const
{
	OP_METHOD (ReturnSensitiveButtons, 0);
	return this->sensitive_buttons;
}


/*
 *	Return TRUE if the control will produce continuous
 *	events while being pressed.
 */

Bool
OPControl::ReturnReportMode () const
{
	OP_METHOD (ReturnReportMode, 0);
	return this->do_cont_reports_on_action;
}



/*
 *	Define the shape style. This does not redraw the
 *	control.
 */

Bool
OPControl::SetShapeStyle (OPCtrlShape shape)
{
	OP_METHOD (SetShapeStyle, 1);
	OP_ARG (1, OPCtrlShape, shape);
	
	this->shape = shape;
	return TRUE;
}



/*
 *	Return the shape style.
 */

OPCtrlShape
OPControl::ReturnShapeStyle () const
{
	OP_METHOD (ReturnShapeStyle, 0);
	
	return this->shape;
}



/*
 *	Define a keyboard alternative. Tell the window about
 *	it...
 */

Bool
OPControl::SetKeyAlt (OPKeyIn key_alt)
{
	OP_METHOD (SetKeyAlt, 1);
	OP_ARG (1, OPKeyIn, key_alt);
	
	if (this->key_alt_known) {
		this->ClearKeyAlt ();
	}
	
	this->key_alt.type      = key_alt.type;
	this->key_alt.value     = key_alt.value;
	this->key_alt.code      = key_alt.code;
	this->key_alt.modifiers = key_alt.modifiers;
	
	this->key_alt_known     = FALSE;
	
	if (this->window.IsValid ()) {
		this->window->NotifyInsertKeyAlt (this);
		this->key_alt_known = TRUE;
	}
	
	return TRUE;
}



/*
 *	Remove a keyboard alternative. Tell the window about
 *	it...
 */

Bool
OPControl::ClearKeyAlt ()
{
	OP_METHOD (ClearKeyAlt, 0);
	
	if (this->key_alt_known) {
		OP_ASSERT (this->window.IsValid ());
		this->window->NotifyRemoveKeyAlt (this);
	}
	
	this->key_alt.type      = OP_KEY_NONE;
	this->key_alt.value     = 0;
	this->key_alt.code      = 0;
	this->key_alt.modifiers = 0;
	
	this->key_alt_known     = FALSE;
	
	return TRUE;
}



/*
 *	Return the keyboard alternative and TRUE if any, FALSE
 *	otherwise.
 */

Bool
OPControl::GetKeyAlt (OPKeyOut key_alt) const
{
	OP_METHOD (GetKeyAlt, 1);
	OP_ARG (1, OPKeyOut, key_alt);
	
	key_alt.type      = this->key_alt.type;
	key_alt.value     = this->key_alt.value;
	key_alt.code      = this->key_alt.code;
	key_alt.modifiers = this->key_alt.modifiers;
	
	return this->key_alt.type != OP_KEY_NONE;
}
	


/*
 *	Return TRUE if a keyboard alternative exists for this
 *	control.
 */

Bool
OPControl::HasKeyAlt () const
{
	OP_METHOD (HasKeyAlt, 0);
	return this->key_alt.type != OP_KEY_NONE;
}

OPKey
OPControl::ReturnKeyAlt () const
{
	OP_METHOD (ReturnKeyAlt, 0);
	return this->key_alt;
}
	


/*
 *	When the control has exported a keyboard alternative
 *	and the window got such a key, the control gets the
 *	opportunity to handle it...
 *
 *	Please override "ActionOnKeyAlt" if the default action
 *	is not what is desired.
 */

Bool
OPControl::HandleKeyAlt (OPEventIn event)
{
	OP_METHOD (HandleKeyAlt, 1);
	OP_ARG (1, OPEventIn, event);
	
	OPKey key = event->ReturnKey ();
	
	key.modifiers &= ~OP_EV_LOCK_MASK;		//	don't look at LOCK keys
	key.modifiers &= ~OP_EV_GEOMOD_MASK;	//	don't matter whether left/right
	
	return ( (this->state & OP_VIEW_IS_ENABLED)
	      && (key.type == this->key_alt.type)
	      && ( (this->key_alt.code && (this->key_alt.code == key.code) )
		    || (this->key_alt.value && (this->key_alt.value == key.value) ) )
	      && (key.modifiers == this->key_alt.modifiers) )
	         ? this->ActionOnKeyAlt ()
			 : FALSE;
}



/*
 *	The user has to override this method in order to
 *	do something when the alternate key gets pressed.
 *	If not, the default action will be issued.
 */

Bool
OPControl::ActionOnKeyAlt ()
{
	OP_METHOD (ActionOnKeyAlt, 0);

	this->ActionOn ();
	this->Display ();
	this->ActionOff ();
	this->Action (FALSE);
	this->Display ();
	
	return TRUE;
}


/*
 *	Draw the active frame around the control. This might be
 *	overridden by some controls which are not compatible with
 *	the very simple default behaviour.
 */

void
OPControl::DrawActiveFrame (OPGrailPortIn port, OPAdornerIn adorner,
							OPCoord margin)
{
	OP_METHOD (DrawActiveFrame, 3);
	OP_ARG (1, OPGrailPortIn, port);
	OP_ARG (2, OPAdornerIn, adorner);
	OP_ARG (3, OPCoord, margin);
	
	adorner->DrawRect (this->box, OP_COLOR_BLACK, margin);
}


/*
 *	Draw the frame of the control. This is has been
 *	implemented in this class, since the frame is used
 *	to show which view is the main responder.
 */

void
OPControl::Draw (OPGrailPortIn port, OPRectIn rect, OPDrawOp what)
{
	OP_METHOD (Draw, 3);
	OP_ARG (1, OPGrailPortIn, port);
	OP_ARG (2, OPRectIn, rect);
	OP_ARG (3, OPDrawOp, what);
	
	const OPAdornerRef& adorner = port->ReturnAdorner ();
	const OPLookManRef& lookman = OPLookMan::Current ();
	
	OPCoord margin = (this->shape & OP_CTRL_NOFRAME) ? 0 : lookman->ReturnWidgetFrameSize ();
	OPCoord effect = (this->shape & OP_CTRL_RELIEF) ? lookman->Return3DEffectWidth () : 0;
	
	if (what & OP_DRAW_BACKGROUND) {
		
		//	If the button is hilited, then draw the button background with the
		//	highlighting color. This means that the superview does not need to
		//	draw the background...
		
		if ( (this->shape & OP_CTRL_AUTOHILITE)
		  && (this->is_being_hilited) ) {
			
			adorner->FillRect (this->box, OP_COLOR_HILITE_CLICK, margin + effect);
			what &= ~OP_DRAW_BACKGROUND;
			
		} else if ( (this->shape & OP_CTRL_AUTOACTHILITE)
				 && (this->is_actively_hilited) ) {
			
			adorner->FillRect (this->box, OP_COLOR_HILITE_ACTIVE, margin + effect);
			what &= ~OP_DRAW_BACKGROUND;
			
		} else if (this->state & OP_VIEW_IS_DROPPING) {
			
//			adorner->FillRect (this->box, OP_COLOR_VIEW_DROPPING, margin + effect);
//			what &= ~OP_DRAW_BACKGROUND;
		}
	}
	
	super::Draw (port, rect, what);
	
	if (what & OP_DRAW_OBJECT) {
		
		if (this->shape & OP_CTRL_RELIEF) {
			
			//	Draw the box around the button. It will be either low (if currently
			//	pressed) or high.
			
			Bool press = (this->shape & OP_CTRL_ABOVE) ? (this->is_being_pressed)
													   : (!this->is_being_pressed);
			
			if (press) {
				adorner->DrawRectLow (this->box, margin);
			} else {
				adorner->DrawRectHigh (this->box, margin);
			}
		}
		
		if ( (this->state & OP_VIEW_IS_ACTIVE)
		  && ((this->shape & OP_CTRL_NOFRAME) == 0)
		  && (this->shape & OP_CTRL_AUTOACTIVE) ) {
			port->SetLineStyle (OP_LINE_DOT);
			this->DrawActiveFrame (port, adorner, margin + effect + 1);
			port->SetLineStyle (OP_LINE_SOLID);
		}
	}
}



/*
 *	Handle some key events which will change the control's
 *	state.
 */

void
OPControl::KeyDown (OPEventIn event)
{	
	OP_METHOD (KeyDown, 1);
	OP_ARG (1, OPEventIn, event);
	
	if ( (event->ReturnType () != OP_EVENT_KEY_DOWN)
	  || (this->is_being_clicked == TRUE) ) {
		return;
	}
	
// ( (event->key_type == OP_KEY_CURSOR) && (event->key_value == OP_CURSOR_UP) )

	if ( (event->ReturnKey ().type == OP_KEY_TAB)
	  && ( (event->ReturnKey ().modifiers & (Card32)(0x0000FFFF)) == 0) ) {
		
		//	Make next view active...
		
		OPViewRef view;
		this->FindNextActive (view);
		this->window->SetMainResponder (view, FALSE);
		
	} else if ( (event->ReturnKey ().type == OP_KEY_TAB)
			 && ( (event->ReturnKey ().modifiers & (Card32)(0x0000FFFF)) == OP_EV_SHIFT) ) {
		
		//	Make previous view active... Since the active list is circular, we
		//	will automatically cycle through every view...
		
		OPViewRef view;
		this->FindPrevActive (view);
		this->window->SetMainResponder (view, FALSE);
		
	} else if ( ( (event->ReturnKey ().type == OP_KEY_RETURN)
			   || (event->ReturnKey ().type == OP_KEY_ENTER)
			   || ( (event->ReturnKey ().type == OP_KEY_ALPHANUM)
				 && (event->ReturnKey ().value == ' ') ) )
			 && (this->state & OP_VIEW_IS_ENABLED) ) {
		
		//	Start an action on this control. This will be activated whenever the
		//	user presses "RETURN", "ENTER" or "SPACE BAR". Usually, "ENTER" will
		//	get to the main "OK" button of the window, if there is one and will
		//	therefore never reach this control !
		
		this->ActionOn ();
		this->Display ();
	}
}



/*
 *	Handle some key events which will change the control's
 *	state.
 */

void
OPControl::KeyUp (OPEventIn event)
{
	OP_METHOD (KeyUp, 1);
	OP_ARG (1, OPEventIn, event);
	
	if ( ( (event->ReturnKey ().type == OP_KEY_RETURN)
	    || (event->ReturnKey ().type == OP_KEY_ENTER)
	    || ( (event->ReturnKey ().type == OP_KEY_ALPHANUM)
	      && (event->ReturnKey ().value == ' ') ) )
	  && ( this->is_being_clicked == TRUE ) ) {
		
		//	End of action... the key being pressed has now been released.
		
		this->ActionOff ();
		this->Action (FALSE);
		this->Display ();
	}
}



/*
 *	Handle a button click with the mouse (begin of click).
 */

void
OPControl::ButtonPress (OPEventIn event)
{
	OP_METHOD (ButtonPress, 1);
	OP_ARG (1, OPEventIn, event);
	
	if ( ((this->state & OP_VIEW_IS_ENABLED) == 0)
	  || (this->is_being_clicked == TRUE) ) {
		return;
	}
	
	if (event->ReturnChange () & this->sensitive_buttons) {
		this->ActionOn ();
		this->state |= OP_VIEW_IS_HOT;
		this->Display ();
	}
	
	//	Avoid reporting dummy presses if the event record does not match.
	//	This might be the case if this method has been invoked directly by
	//	the user, not by the general event dispatching process.
	
	if (event->ReturnType () == OP_EVENT_PRESS) {
		
		Card32 button = event->ReturnChange ();
		
		switch (button) {
			case OP_EV_BUTTON_LEFT:		this->Report ("LeftButtonPress");	break;
			case OP_EV_BUTTON_RIGHT:	this->Report ("RightButtonPress");	break;
			case OP_EV_BUTTON_MIDDLE:	this->Report ("MiddleButtonPress");	break;
			case OP_EV_BUTTON_EXTRA:	this->Report ("ExtraButtonPress");	break;
		}
	}
}



/*
 *	Handle a button click with the mouse (end of click).
 */

void
OPControl::ButtonRelease (OPEventIn event)
{
	OP_METHOD (ButtonRelease, 1);
	OP_ARG (1, OPEventIn, event);
	
	//	Avoid reporting dummy releases if the event record does not match.
	//	This might be the case if this method has been invoked directly by
	//	the user, not by the general event dispatching process.
	
	if ( (this->state & OP_VIEW_IS_ENABLED)
	  && (event->ReturnType () == OP_EVENT_RELEASE) ) {
		
		Card32 button = event->ReturnChange ();
		
		switch (button) {
			case OP_EV_BUTTON_LEFT:		this->Report ("LeftButtonRelease");		break;
			case OP_EV_BUTTON_RIGHT:	this->Report ("RightButtonRelease");	break;
			case OP_EV_BUTTON_MIDDLE:	this->Report ("MiddleButtonRelease");	break;
			case OP_EV_BUTTON_EXTRA:	this->Report ("ExtraButtonRelease");	break;
		}
	}
	
	if ( ((this->state & OP_VIEW_IS_ENABLED) == 0)
	  || (this->is_being_clicked == FALSE) ) {
		return;
	}
	
	if (event->ReturnChange () & this->sensitive_buttons) {
		this->state &= ~ OP_VIEW_IS_HOT;
		this->ActionOff ();
		this->Action (TRUE);
		this->Display ();
	}
}



void
OPControl::EnterNotify (OPEventIn event)
{
	OP_METHOD (EnterNotify, 1);
	OP_ARG (1, OPEventIn, event);
	
	this->Report ("EnterNotify");
	
	if ( ((this->state & OP_VIEW_IS_ENABLED) == 0)
	  || (this->shape & OP_CTRL_NOTACTENTER)
	  || (this->is_being_clicked == TRUE) ) {
		return;
	}
	
	if (event->ReturnButtons () & this->sensitive_buttons) {
		this->ActionOn ();
		this->Display ();
	}
}



void
OPControl::LeaveNotify (OPEventIn event)
{
	OP_METHOD (LeaveNotify, 1);
	OP_ARG (1, OPEventIn, event);
	
	this->Report ("LeaveNotify");
	
	if ( ((this->state & OP_VIEW_IS_ENABLED) == 0)
	  || (this->is_being_clicked == FALSE) ) {
		return;
	}
	
	if (event->ReturnButtons () & this->sensitive_buttons) {
		this->ActionOff ();
		this->Display ();
	}
}



/*
 *	The control is being pressed and continuous events
 *	will be sent.
 */

void
OPControl::Tick ()
{
	OP_METHOD (Tick, 0);
	
	static int recursion = 0;
	
	recursion++;
	
	if (recursion == 1) {
		if (this->is_app_ticking) {
			this->app->InsertTickResponder (this, CONTINUOUS_MS);
		}
	}
	
	this->Report (notify_OnContClick, OP_REPORT_NEW);
	
	recursion--;
}


/*
 *	Set the minimum size.
 */

Bool
OPControl::SetDesiredMinSize (OPCoord dx, OPCoord dy)
{
	OP_METHOD (SetDesiredMinSize, 2);
	OP_ARG (1, OPCoord, dx);
	OP_ARG (2, OPCoord, dy);
	
	this->min_dx = dx;
	this->min_dy = dy;
	
	return TRUE;
}

Bool
OPControl::GetDesiredMinSize (OPCoordOut dx, OPCoordOut dy) const
{
	OP_METHOD (GetDesiredMinSize, 2);
	OP_ARG (1, OPCoordOut, dx);
	OP_ARG (2, OPCoordOut, dy);
	
	dx = this->min_dx;
	dy = this->min_dy;
	
	return TRUE;
}


/*
 *	Return the control's minimal size (depends both on
 *	the fixed minimum and on the super-view's minimum).
 */

void
OPControl::GetMinSize (OPCoordOut dx, OPCoordOut dy) const
{
	OP_METHOD (GetMinSize, 2);
	OP_ARG (1, OPCoordOut, dx);
	OP_ARG (2, OPCoordOut, dy);
	
	OPCoord dx1 = 0;
	OPCoord dy1 = 0;
	OPCoord dx2 = this->min_dx;
	OPCoord dy2 = this->min_dy;
	
	super::GetMinSize (dx1, dy1);
	
	dx = dx1 > dx2 ? dx1 : dx2;
	dy = dy1 > dy2 ? dy1 : dy2;
}


/*
 *	Return the control's maximum size. Return an arbitrary
 *	huge value.
 */

void
OPControl::GetMaxSize (OPCoordOut dx, OPCoordOut dy) const
{
	OP_METHOD (GetMinSize, 2);
	OP_ARG (1, OPCoordOut, dx);
	OP_ARG (2, OPCoordOut, dy);
	
	super::GetMaxSize (dx, dy);
}





/*
 *	Broadcast some message...
 */

void
OPControl::Broadcast (OPAtomIn name, OPObjectIn param)
{
	OP_METHOD (Broadcast, 2);
	OP_ARG (1, OPAtomIn, name);
	OP_ARG (1, OPObjecIn, param);
	
	super::Broadcast (name, param);
}

