/*
 *	dictree.cxx
 *
 *	The Dictionary Tree is a simple implementation of a dictionary
 *	like tree structure which can be accessed using paths instead of
 *	simple keys.
 *
 *	(C) Copyright 1994-1998, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#include <OPaC/dictree.h>
#include <OPaC/iterator.h>
#include <OPaC/new.h>
#include <OPaC/string.h>

#include <string.h>

/****************************************************************/

OP_IMPLEMENT_CLASS (OPDicTree);

/****************************************************************/


/*
 *	Create an empty dictionary tree.
 */

OPDicTree::OPDicTree ()
{
	OP_METHOD (OPDicTree, 0);
	
	this->root = OP_NEW_FROM_ZONE (this->ReturnMemZone (), OPDict);
	this->root->SetAttribute (this, OPObjectNull);
}



/*
 *	Destroy the dictionary tree.
 */

OPDicTree::~OPDicTree ()
{
	OP_METHOD (~OPDicTree, 0);
	
	this->root->RemoveAll ();
	this->root->RemoveAttribute (this);
}



/*
 *	Save the dictionary tree to the storage.
 */

void
OPDicTree::StoreInstance (OPStorageIn storage)
{
	OP_METHOD (StoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	super::StoreInstance (storage);
	
	storage << this->root;
	storage << OPStorageSync::mark;
}



/*
 *	Restore the dictionary from the storage.
 */

void
OPDicTree::RestoreInstance (OPStorageIn storage)
{
	OP_METHOD (RestoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	super::RestoreInstance (storage);
	
	storage >> this->root;
	storage >> OPStorageSync::mark;
}



/*
 *	Prepare the dictionary tree for its destruction.
 */

void
OPDicTree::Delete ()
{
	OP_METHOD (Delete, 0);
	
	this->RemoveAll ();
	super::Delete ();
}

OPIteratorRef
OPDicTree::NewIterator ()
{
	OPDicTreeIteratorRef iterator = OP_NEW (OPDicTreeIterator);
	
	this->InsertIterator (iterator);
	this->iterators->Insert (iterator);
	return iterator;
}


/*
 *	Find the root directory by analysing the path.
 *
 *	Return :
 *
 *	- TRUE => simple key to be used directly in the output
 *	  dictionary.
 *	- FALSE => key is still a path with a root starting at
 *	  the output dictionary.
 */

Bool
OPDicTree::FindRootDict (OPObjectIn path, OPObjectOut key, OPDictOut dict)
{
	OP_METHOD (FindRootDict, 3);
	OP_ARG (1, OPObjectIn, path);
	OP_ARG (2, OPObjectOut, key);
	OP_ARG (3, OPDictOut, dict);
	
	key  = path;
	dict = this->root;
	
	if (path.IsValid ()) {
		if (path->IsKindOf (OP_RUNTIME (OPGenericContainer))) {
			
			OPGenericContainerRef cont = path;
			OPObjectRef subkey;
			OPObjectRef subdict;
			OPObjectRef data;
			Count index = 0;
			Count total = cont->TotalCount ();
			
			while (index < total) {
				
				cont->Get (index, subkey);
				
				if (dict->GetFromKey (subkey, subdict) == FALSE) break;
				if (subdict.IsInvalid ()) break;
				if (subdict->IsKindOf (OP_RUNTIME (OPDict)) == FALSE) break;
				if (subdict->GetAttribute (this, data) == FALSE) break;
				
				//	Move one level deeper into the tree.
				
				dict = subdict;
				index++;
			}
			
			if (index == total) {
				
				//	We reached a leaf. The leaf is in fact a node...
				
				key = OPObjectNull;
				
			} else if (index+1 == total) {
				
				//	We reached a node and there the remaining key points to the
				//	final leaf.
				
				cont->Get (index, key);
				
			} else {
				
				//	We had to stop because the requested dictionary node could
				//	not be found. The key is the partial path starting from the
				//	last valid node.
				
				OPListRef list = OP_NEW (OPList);
				
				while (index < total) {
					cont->Get (index, subkey);
					list->PushTail (subkey);
					index++;
				}
				
				key = list;
				
				return FALSE;
			}
		}
	}
	
	return TRUE;
}

/*
 *	Insert a key/value pair into the dictionary tree.
 */

Bool
OPDicTree::Insert (OPObjectIn key, OPObjectIn value)
{
	OP_METHOD (Insert, 2);
	OP_ARG (1, OPObjectIn, key);
	OP_ARG (2, OPObjectIn, value);
	
	OPDictRef dict;
	OPObjectRef subkey;
	
	if (this->FindRootDict (key, subkey, dict)) {
		if (subkey.IsInvalid ()) {
			return FALSE;
		} else {
			if (dict->Remove (subkey)) {
				this->num_element--;
			}
			if (dict->Insert (subkey, value)) {
				this->num_element++;
			}
			return TRUE;
		}
	}
	
	OPListRef list = subkey;
	
	while (list->TotalCount () > 1) {
		OPDictRef subdict = OP_NEW_FROM_ZONE (this->ReturnMemZone (), OPDict);
		list->PopHead (subkey);
		subdict->SetAttribute (this, dict);
		dict->Insert (subkey, subdict);
		dict = subdict;
	}
	
	list->PopHead (subkey);
	dict->Insert (subkey, value);
	this->num_element++;
	
	return FALSE;
}



/*
 *	Remove an element from the dictionary tree. Indexes are not
 *	supported (yet).
 */

Bool
OPDicTree::Remove (Count pos)
{
	OP_METHOD (Remove, 1);
	OP_ARG (1, Count, pos);
	
	return FALSE;
}



/*
 *	Remove an object specified by its key.
 */

Bool
OPDicTree::Remove (OPObjectIn key, Count nth)
{
	OP_METHOD (Remove, 2);
	OP_ARG (1, OPObjectIn, key);
	OP_ARG (2, Count, nth);
	
	OPDictRef dict;
	OPObjectRef subkey;
	
	if (this->FindRootDict (key, subkey, dict)) {
		if (subkey.IsInvalid ()) {
			
			//	Remove a branch of the tree.
			
			OPDictRef parent;
			dict->GetAttribute (this, parent);
			this->RemoveSubTree (dict);
			return parent->RemoveValue (dict);
			
		} else {
			
			if (dict->Remove (subkey, nth)) {
				this->num_element--;
				return TRUE;
			}
		}
	}
	
	return FALSE;
}


void
OPDicTree::RemoveSubTree (OPDictIn dict)
{
	OP_METHOD (RemoveSubTree, 1);
	OP_ARG (1, OPDictIn, dict);
	
	OPListRef list = OP_NEW (OPList);
	list->PushTail (dict);
	
	do {
		
		//	Examine each level in turn. First start at level 0, then
		//	put the next levels into the examination list, then the next,
		//	then the next, etc.
		
		OPDictRef dict;
		Count index = 0;
		list->PopTail (dict);
		
		OPObjectRef key;
		OPObjectRef val;
		OPObjectRef data;
		
		while (dict->Get (index++, key, val)) {
			if ( (val.IsValid ())
			  && (val->IsKindOf (OP_RUNTIME (OPDict)))
			  && (val->GetAttribute (this, data)) ) {
				list->PushTail (val);
			} else {
				this->num_element--;
			}
		}
		
		dict->RemoveAll ();
		dict->RemoveAttribute (this);
		dict = 0;
		
	} while (list->TotalCount ());
}


/*
 *	Remove every object from the dictionary tree.
 */

void
OPDicTree::RemoveAll ()
{
	OP_METHOD (RemoveAll, 0);
	
	this->RemoveSubTree (this->root);
	this->root->SetAttribute (this, OPObjectNull);
	this->num_element = 0;
}

Bool
OPDicTree::RemoveValue (OPObjectIn v, Count nth)
{
	OP_METHOD (RemoveValue, 2);
	OP_ARG (1, OPObjectIn, v);
	OP_ARG (2, Count, nth);
	
	//	NIY
	
	return FALSE;
}


/*
 *	Find the nth leaf in the tree, starting at the specified dictionary
 *	node (this skips "pos" leaves). Either return TRUE and the valid node
 *	and index corresponding to the nth leaf or return FALSE and the number
 *	of leaves which must still be skipped.
 *
 *	The path is updated on success only...
 */

Bool
OPDicTree::FindSubTree (OPDictIn dict, Count pos,
						OPDictOut found, CountOut index, OPListIn path)
{
	OP_METHOD (FindSubTree, 5);
	OP_ARG (1, OPDictIn, dict);
	OP_ARG (2, Count, pos);
	OP_ARG (3, OPDictOut, found);
	OP_ARG (4, CountOut, index);
	OP_ARG (5, OPListIn, path);
	
	Count       num  = 0;
	OPDictRef   iter = dict;
	OPObjectRef key;
	OPObjectRef val;
	OPObjectRef data;
	
	while (iter->Get (num, key, val)) {
		if ( (val.IsValid ())
		  && (val->IsKindOf (OP_RUNTIME (OPDict)))
		  && (val->GetAttribute (this, data)) ) {
			
			//	Get deeper into the subtree. If we found what we were looking for,
			//	return TRUE and the result. Otherwise, skip as many elements as we
			//	found while traversing the full subtree.
			
			if (path.IsValid ()) path->PushTail (key);
			
			if (this->FindSubTree (val, pos, found, index, path)) {
				return TRUE;
			}
			
			if (path.IsValid ()) path->PopTail (key);
			pos = index;
			
		} else if (pos == 0) {
			
			//	Reached the leaf we were looking for => return the index of the leaf
			//	in the deepest node.
			
			found = iter;
			index = num;
			
			return TRUE;
			
		} else {
			
			//	We have just skipped a leaf...
			
			pos--;
		}
		
		num++;
	}
	
	found = OPDictNull;
	index = pos;
	
	return FALSE;
}

/*
 *	Find the key/value pair based on an index in the
 *	dictionary tree. This is not implemented (yet).
 */

Bool
OPDicTree::Get (Count pos, OPObjectOut key, OPObjectOut value)
{
	OP_METHOD (Get, 3);
	OP_ARG (1, Count, pos);
	OP_ARG (2, OPObjectOut, key);
	OP_ARG (3, OPObjectOut, value);
	
	OPDictRef dict;
	OPListRef path = OP_NEW (OPList);
	
	if (this->FindSubTree (this->root, pos, dict, pos, path)) {
		if (dict->Get (pos, key, value)) {
			path->PushTail (key);
			key = path;
			return TRUE;
		}
	}
	
	return FALSE;
}



/*
 *	Find the value of the nth corresponding key. Return
 *	FALSE on failure.
 */

Bool
OPDicTree::GetFromKey (OPObjectIn key, OPObjectOut value, Count nth)
{
	OP_METHOD (GetFromKey, 3);
	OP_ARG (1, OPObjectIn, key);
	OP_ARG (2, OPObjectOut, value);
	OP_ARG (3, Count, nth);
	
	OPDictRef dict;
	OPObjectRef subkey;
	
	if (this->FindRootDict (key, subkey, dict)) {
		if (subkey.IsInvalid ()) {
			if (nth == 0) {
				value = dict;
				return TRUE;
			}
		} else {
			return dict->GetFromKey (subkey, value, nth);
		}
	}
	
	return FALSE;
}



/*
 *	Find the key of the nth corresponding value. Return
 *	FALSE on failure. This is not implemented yet.
 */

Bool
OPDicTree::GetFromValue (OPObjectIn value, OPObjectOut key, Count nth)
{
	OP_METHOD (GetFromValue, 3);
	OP_ARG (1, OPObjectIn, value);
	OP_ARG (2, OPObjectOut, key);
	OP_ARG (3, Count, nth);
	
	return FALSE;
}

/****************************************************************/

Bool
OPDicTree::SplitTextPath (const LatinChar* path, OPListOut list)
{
	OP_METHOD (SplitTextPath, 2);
	OP_ARG (1, const LatinChar*, path);
	OP_ARG (2, OPListOut, list);
	
	if ( (path == 0)
	  || (path[0] == 0) ) {
		return FALSE;
	}
	
	Count length = strlen (path);
	Count numdir = 0;
	Count index  = 0;
	
	LatinChar* buffer = (LatinChar*) stack_alloc (length+1);
	LatinChar* name = buffer;
	
	strcpy (name, path);
	
	list = OP_NEW (OPList);
	
	//	Cut the name into pieces and build the path.
	
	for (index = 0; index < length; index++) {
		if (buffer[index] == ':') {
			buffer[index] = 0;
			OPAtomRef part = name;
			list->PushTail (part);
			name = buffer + index + 1;
		}
	}
	
	OPAtomRef part = name;
	list->PushTail (part);
	
	return TRUE;
}

Bool
OPDicTree::InsertUsingPath (OPStringIn path, OPObjectIn data)
{
	OP_METHOD (InsertUsingPath, 2);
	OP_ARG (1, OPStringIn, path);
	OP_ARG (2, OPObjectIn, data);
	
	if (path.IsValid ()) {
		Count length = path->CountCharacters ();
		LatinChar* buffer = (LatinChar*) stack_alloc (length+1);
		path->GetStringCopy (buffer, length);
		return this->InsertUsingPath (buffer, data);
	}
	
	return FALSE;
}

Bool
OPDicTree::InsertUsingPath (const LatinChar* path, OPObjectIn data)
{
	OP_METHOD (InsertUsingPath, 2);
	OP_ARG (1, const LatinChar*, path);
	OP_ARG (2, OPObjectIn, data);
	
	OPListRef key;
	
	if (this->SplitTextPath (path, key)) {
		return this->Insert (key, data);
	}
	
	return FALSE;
}

/****************************************************************/

Bool
OPDicTree::GetFromPath (OPStringIn path, OPObjectOut data)
{
	OP_METHOD (GetFromPath, 2);
	OP_ARG (1, OPStringIn, path);
	OP_ARG (2, OPObjectIn, data);
	
	if (path.IsValid ()) {
		Count length = path->CountCharacters ();
		LatinChar* buffer = (LatinChar*) stack_alloc (length+1);
		path->GetStringCopy (buffer, length);
		return this->GetFromPath (buffer, data);
	}
	
	return FALSE;
}

Bool
OPDicTree::GetFromPath (const LatinChar* path, OPObjectOut data)
{
	OP_METHOD (GetFromPath, 2);
	OP_ARG (1, const LatinChar*, path);
	OP_ARG (2, OPObjectIn, data);
	
	OPListRef key;
	
	if (this->SplitTextPath (path, key)) {
		return this->GetFromKey (key, data);
	}
	
	return FALSE;
}

/****************************************************************/
