/*
 *	filewr.cxx
 *
 *	The FileWrite class is a specialized stream used to write data to a file.
 *
 *	(C) Copyright 1994-1998, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#include <OPaC/string.h>
#include <OPaC/stream.h>
#include <OPaC/new.h>

/****************************************************************/

OP_IMPLEMENT_CLASS (OPFileWrite);

/****************************************************************/

#define	MAX_FILENAME_LEN	300


/*
 *	The File Write Stream provides an interface for writing
 *	data sequentially to a file.
 *
 *	The file can either be created from scratch or just be
 *	expanded from its previous state.
 */

OPFileWrite::OPFileWrite ()
{
	OP_METHOD (OPFileWrite, 0);
	
	this->file = 0;
}



/*
 *	Close the file if still open at destruction time.
 */

OPFileWrite::~OPFileWrite ()
{
	OP_METHOD (~OPFileWrite, 0);
	
	if (this->file) {
		OPaC_CloseFile (this->file);
	}
}



/*
 *	Don't store anything yet.
 */
	
void
OPFileWrite::StoreInstance (OPStorageIn storage)
{
	OP_METHOD (StoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	super::StoreInstance (storage);
	
	//	NIY
}



/*
 *	Don't restore anything yet.
 */

void
OPFileWrite::RestoreInstance (OPStorageIn storage)
{
	OP_METHOD (RestoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	super::RestoreInstance (storage);
	
	//	NIY
}



/*
 *	If append is FALSE :
 *
 *	- Open a file from scratch. This will erase its previous
 *	  contents.
 *
 *	If append is TRUE :
 *
 *	- Open a file to append data to it. This will expand the
 *	  file from its current end.
 *
 *	If another file was under use before, close it. The file
 *	name has to be a physical, OS dependent file name !
 *
 *	Throws :
 *
 *	- OPExcStream (CANNOT_OPEN) if an error occurred while
 *	  opening the file.
 */

void
OPFileWrite::OpenWrite (const LatinChar* name, Bool append)
{
	OP_METHOD (OpenWrite, 2);
	OP_ARG (1, const LatinChar*, name);
	OP_ARG (2, Bool, append);
	
	OPStringRef string = name;
	this->OpenWrite (string, append);
}

void
OPFileWrite::OpenWrite (OPStringIn name, Bool append)
{
	OP_METHOD (OpenWrite, 2);
	OP_ARG (1, OPStringIn, name);
	OP_ARG (2, Bool, append);
	
	if (this->file) {
		this->Close ();
	}
	
	this->file = (append) ? OPaC_OpenFileAppend (name)
						  : OPaC_OpenFileWrite (name);
	
	if (this->file == 0) {
		THROW (OPExcStream (OPExcStream::CANNOT_OPEN));
	}
}


/*
 *	Close the current file. This will flush the data still
 *	remaining in the internal buffers.
 *
 *	Throws :
 *
 *	- OPExcStream (WRITE_ERROR) if an error occurred while
 *	  closing the file.
 */

void
OPFileWrite::Close ()
{
	OP_METHOD (Close, 0);
	
	if (this->file) {
		if (OPaC_CloseFile (this->file) == FALSE) {
			this->file = 0;
			THROW (OPExcStream (OPExcStream::WRITE_ERROR));
		}
		
		this->file = 0;
	}
}



/*
 *	Write data to the file.
 *
 *	Throws :
 *
 *	- OPExcStream (WRITE_ERROR) if an error occurred while
 *	  writing to the file.
 *
 *	- OPExcStream (NO_WRITE) if no file has been opened.
 */

Count
OPFileWrite::Write (const void* buffer, Count size)
{
	OP_METHOD (Write, 2);
	OP_ARG (1, const void*, buffer);
	OP_ARG (2, Count, size);
	
	if (this->file == 0) {
		THROW (OPExcStream (OPExcStream::NO_WRITE));
	}
	
	if (OPaC_WriteFile (this->file, buffer, size) != size) {
		THROW (OPExcStream (OPExcStream::WRITE_ERROR));
	}

	return size;
}



/*
 *	Synchronize the file with the stream : this flushes the
 *	internal buffers.
 *
 *	Throws :
 *
 *	- OPExcStream (WRITE_ERROR) if an error occurred while
 *	  synchronizing the file.
 *
 *	- OPExcStream (NO_WRITE) if no file has been opened.
 */

void
OPFileWrite::Sync ()
{
	OP_METHOD (Sync, 0);
	
	if (this->file == 0) {
		THROW (OPExcStream (OPExcStream::NO_WRITE));
	}
	
	if (OPaC_FlushFile (this->file) == FALSE) {
		THROW (OPExcStream (OPExcStream::WRITE_ERROR));
	}
}



/*
 *	Truncate file to the current position.
 *
 *	Throws :
 *
 *	- OPExcStream (NO_WRITE) if no file has been opened.
 */

void
OPFileWrite::Trunc ()
{
	OP_METHOD (Trunc, 0);
	
	if (this->file == 0) {
		THROW (OPExcStream (OPExcStream::NO_WRITE));
	}
	
	OPaC_TruncFile (this->file);
}

