/*
 *	list.cxx
 *
 *	The OPaC List class provides a sorted list in which objects
 *	can be stored. Their position depends on their contents.
 *
 *	(C) Copyright 1994-1997, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#include <OPaC/object.h>
#include <OPaC/container.h>
#include <OPaC/listit.h>
#include <OPaC/new.h>

/****************************************************************/

OP_IMPLEMENT_CLASS (OPList);

/****************************************************************/


/*
 *	Create an empty list, which will by default be sorted.
 */

OPList::OPList ()
{
	OP_METHOD (OPList, 0);
	
	this->is_sorted      = TRUE;
	this->sort_on_insert = TRUE;
	
	this->head = 0;
	this->tail = 0;
	
	this->cache_elem  = 0;
	this->cache_index = 0;
}


/*
 *	Destroy the list. Every element maintained in the list
 *	will be removed. If no longer referenced, these will
 *	automagically be deleted.
 */

OPList::~OPList ()
{
	OP_METHOD (~OPList, 0);
	
	if (this->num_element) {
		this->RemoveAll ();
	}
}



/*
 *	Save a list to an archive.
 */

void
OPList::StoreInstance (OPStorageIn storage)
{
	OP_METHOD (StoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	super::StoreInstance (storage);
	
	storage << this->is_sorted << this->sort_on_insert;
	storage << OPStorageSync::mark;
	
	Element* elem = this->head;
	
	while (elem) {
		storage << elem->data;
		elem = elem->next;
	}
	
	storage << OPStorageSync::mark;
}



/*
 *	Restore a list from an archive.
 */

void
OPList::RestoreInstance (OPStorageIn storage)
{
	OP_METHOD (RestoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	super::RestoreInstance (storage);
	
	storage >> this->is_sorted >> this->sort_on_insert;
	storage >> OPStorageSync::mark;
	
	Count num = this->num_element;
	
	while (num--) {
		
		Element* elem = NEW_FROM_ZONE (this->ReturnMemZone ()) Element;
		
		elem->next = 0;
		elem->prev = this->tail;
		elem->data = 0;
		
		(elem->prev ? elem->prev->next : this->head) = elem;
		(elem->next ? elem->next->prev : this->tail) = elem;
		
		storage >> elem->data;
	}
	
	storage >> OPStorageSync::mark;
	
	this->cache_elem  = 0;
	this->cache_index = 0;
}



/*
 *	Prepare the list for its destruction.
 */

void
OPList::Delete ()
{
	OP_METHOD (Delete, 0);
	
	this->RemoveAll ();
	super::Delete ();
}


OPIteratorRef
OPList::NewIterator ()
{
	OPListIteratorRef iterator = OP_NEW (OPListIterator);
	
	this->InsertIterator (iterator);
	this->iterators->Insert (iterator);
	return iterator;
}



/*
 *	Insert an object into the list. If the list is defined
 *	as sorted, then insert at the right place.
 *
 *	Return TRUE if successful. Invalid objects will not be
 *	accepted.
 */

Bool
OPList::Insert (OPObjectIn value)
{
	OP_METHOD (Insert, 1);
	OP_ARG (1, OPObjectIn, value);
	
	if (value.IsValid () == FALSE) {
		return FALSE;
	}
	
	Element* elem  = NEW_FROM_ZONE (this->ReturnMemZone ()) Element;
	
	if (this->sort_on_insert) {
		
		//	If the list is sorted, we will do a sorted insertion.
		
		Element* find  = this->head;
		Count    index = 0;
		
		//	Find the element which will become the successor in the sorted list...
		//	This is quite ineffective - should try to use splay or Red-Black trees.
		
		while (find) {
			
			if (find->data->CompareWith (value) == OP_COMP_THIS_AFTER) {
				break;
			}
			
			find = find->next;
			index++;
		}
		
		elem->next = (find) ? find : 0;
		elem->prev = (find) ? find->prev : this->tail;
		
		if ( (this->cache_elem)
		  && (this->cache_index >= index) ) {
			
			this->cache_index++;
		}
		
	} else {
		
		//	If the list is no longer sorted, then just insert at the end of the
		//	list...
		
		elem->next = 0;
		elem->prev = this->tail;
		
		this->is_sorted = FALSE;
	}
	
	elem->data = value;
	(elem->prev ? elem->prev->next : this->head) = elem;
	(elem->next ? elem->next->prev : this->tail) = elem;
	this->num_element++;
	
	if (this->iterators.IsValid ()) {
		Count index = 0;
		OPListIteratorRef iter;
		
		while (this->iterators->Get (index++, iter)) {
			if (iter.IsValid ()) {
				iter->NotifyInsert (elem);
			}
		}
	}
	
	return TRUE;
}



/*
 *	Remove the nth object which is equal to the specified
 *	one. Equality means same pointers.
 *
 *	Return TRUE if successful.
 */

Bool
OPList::Remove (OPObjectIn value, Count nth)
{
	OP_METHOD (Remove, 2);
	OP_ARG (1, OPObjectIn, value);
	OP_ARG (2, Count, nth);
	
	Element* elem = this->head;
	Bool found_before = FALSE;
	
	while (elem) {
		
		if ( (elem->data == value)
		  && (nth-- == 0) ) {
			
			if (this->iterators.IsValid ()) {
				Count index = 0;
				OPListIteratorRef iter;
				
				while (this->iterators->Get (index++, iter)) {
					if (iter.IsValid ()) {
						iter->NotifyRemove (elem);
					}
				}
			}
			
			if (this->cache_elem == elem) {
				this->cache_elem  = 0;
				this->cache_index = 0;
			}
			if ( (this->cache_elem)
			  && (found_before == FALSE) ) {
				this->cache_index--;
			}
			
			(elem->next ? elem->next->prev : this->tail) = elem->prev;
			(elem->prev ? elem->prev->next : this->head) = elem->next;
			this->num_element--;
			
			elem->data = 0;
			
			DELETE (elem);
			
			return TRUE;
		}
		
		if (elem == this->cache_elem) {
			found_before = TRUE;
		}
		
		elem = elem->next;
	}
	
	return FALSE;
}



/*
 *	Remove an object defined by its position in the list.
 *	Return TRUE if successful.
 */

Bool
OPList::Remove (Count pos)
{
	OP_METHOD (Remove, 1);
	OP_ARG (1, Count, pos);
	
	if (pos >= this->num_element) {
		return FALSE;
	}
	
	Element* elem = this->head;
	Count    skip = pos;
	
	while (skip--) {
		elem = elem->next;
	}
	
	if (this->cache_elem == elem) {
		this->cache_elem  = 0;
		this->cache_index = 0;
	}
	
	if ( (this->cache_elem)
	  && (this->cache_index >= pos) ) {
		this->cache_index--;
	}
	
	if (this->iterators.IsValid ()) {
		Count index = 0;
		OPListIteratorRef iter;
		
		while (this->iterators->Get (index++, iter)) {
			if (iter.IsValid ()) {
				iter->NotifyRemove (elem);
			}
		}
	}
	
	(elem->next ? elem->next->prev : this->tail) = elem->prev;
	(elem->prev ? elem->prev->next : this->head) = elem->next;
	this->num_element--;
	OPObjectRef& value = elem->data;
	value = 0;
	DELETE (elem);
	
	return TRUE;
}



/*
 *	Remove every object contained in the list.
 */

void
OPList::RemoveAll ()
{
	OP_METHOD (RemoveAll, 0);
	
	while (this->head) {
		Element*     elem  = this->head;
		OPObjectRef& value = elem->data;
		
		if (this->iterators.IsValid ()) {
			Count index = 0;
			OPListIteratorRef iter;
			
			while (this->iterators->Get (index++, iter)) {
				if (iter.IsValid ()) {
					iter->NotifyRemove (elem);
				}
			}
		}
		
		this->head = elem->next;
		value = 0;
		DELETE (elem);
	}
	
	this->head        = 0;
	this->tail        = 0;
	this->is_sorted   = TRUE;
	this->num_element = 0;
}



/*
 *	Find the object at the specified position in the list.
 *	This will return TRUE if successful.
 */

Bool
OPList::Get (Count pos, OPObjectOut object)
{
	OP_METHOD (Get, 2);
	OP_ARG (1, Count, pos);
	OP_ARG (2, OPObjectOut, object);
	
	if (pos >= this->num_element) {
		return FALSE;
	}
	
	Element* elem = this->head;
	Count    skip = pos;
	
	if ( (this->cache_elem)
	  && (this->cache_index <= skip) ) {
		
		elem = this->cache_elem;
		skip = skip - this->cache_index;
	}
	
	while (skip--) {
		elem = elem->next;
	}
	
	this->cache_elem  = elem;
	this->cache_index = pos;
	
	object = elem->data;
	
	return TRUE;
}



/* 
 *	Look for some similar object in the list (take the nth,
 *	if nth == 0 take the head one). Comparison is based
 *	on the "CompareWith" method. Return TRUE if successful.
 */

Bool
OPList::GetSimilar (OPObjectIn value, OPObjectOut object, Count nth)
{
	OP_METHOD (GetSimilar, 3);
	OP_ARG (1, OPObjectIn, value);
	OP_ARG (2, OPObjectOut, object);
	OP_ARG (3, Count, nth);
	
	Element* elem = this->head;
	
	while (elem) {
		
		if ( (elem->data->CompareWith (value) == OP_COMP_THIS_SAME)
		  && (nth-- == 0) ) {
			
			object = elem->data;
			return TRUE;
		}
		
		elem = elem->next;
	}
	
	return FALSE;
}



/*
 *	Return TRUE if the object can be found in the list.
 *	The comparison is based on the pointers.
 */

Bool
OPList::LookFor (OPObjectIn value)
{
	OP_METHOD (LookFor, 1);
	OP_ARG (1, OPObjectIn, value);
	
	Element* elem = this->head;
	
	while (elem) {
		
		if (elem->data == value) {
			return TRUE;
		}
		
		elem = elem->next;
	}
	
	return FALSE;
}


/*
 *	Return TRUE if the object can be found in the list,
 *	starting at position "index". The index will be
 *	updated to match the position where the object was
 *	found.
 */

Bool
OPList::Find (OPObjectIn value, CountInOut index)
{
	OP_METHOD (Find, 2);
	OP_ARG (1, OPObjectIn, value);
	OP_ARG (2, CountInOut, index);
	
	Element* elem = 0;
	
	return this->Find (value, index, elem);
}

Bool
OPList::Find (OPObjectIn value, CountInOut index, Element* elem)
{
	OP_METHOD (Find, 3);
	OP_ARG (1, OPObjectIn, value);
	OP_ARG (2, CountInOut, index);
	OP_ARG (3, Element*, elem);
	
	elem = this->head;
	Count skip = index;
	
	if ( (this->cache_elem)
	  && (this->cache_index <= skip) ) {
		
		elem = this->cache_elem;
		skip = skip - this->cache_index;
	}
	
	while (elem) {
		if (skip) {
			skip--;
		} else {
			if (elem->data == value) {
				this->cache_elem  = elem;
				this->cache_index = index;
				return TRUE;
			}
			index++;
		}
		elem = elem->next;
	}
	
	return FALSE;
}


/*
 *	Move an (existing) element in the list.
 */

Bool
OPList::MoveToHead (OPObjectIn value)
{
	OP_METHOD (MoveToHead, 1);
	OP_ARG (1, OPObjectIn, value);
	
	return this->MoveAtPosition (value, 0);
}

Bool
OPList::MoveToTail (OPObjectIn value)
{
	OP_METHOD (MoveToHead, 1);
	OP_ARG (1, OPObjectIn, value);
	
	return this->MoveAtPosition (value, this->num_element);
}


/*
 *	Move an (existing) element in the list. The position
 *	is specified absolutely, before the object is removed
 *	from the list; moving an object from pos to pos or pos+1
 *	won't move it at all.
 */

Bool
OPList::MoveAtPosition (OPObjectIn value, Count pos)
{
	OP_METHOD (MoveAtPosition, 2);
	OP_ARG (1, OPObjectIn, value);
	OP_ARG (2, Count, pos);
	
	if (this->sort_on_insert) {
		return FALSE;
	}
	
	if (pos > this->num_element) {
		pos = this->num_element;
	}
	
	Count    index = 0;
	Element* elem  = 0;
	
	if (this->Find (value, index, elem)) {
		
		if ( (pos == index)
		  || (pos == index+1) ) {
			return TRUE;
		}
		
		this->is_sorted = FALSE;
		
		if (pos < index) {
			
			//	Find the interesting element.
			
			Element* iter = this->head;
			Count    skip = pos;
			
			while (iter && skip) {
				iter = iter->next;
				skip--;
			}
			
			//	Remove at current position.
			
			(elem->next ? elem->next->prev : this->tail) = elem->prev;
			(elem->prev ? elem->prev->next : this->head) = elem->next;
			
			//	Insert before "iter".
			
			elem->prev = iter->prev;
			elem->next = iter;
			
			(iter->prev ? iter->prev->next : this->head) = elem;
			iter->prev = elem;
			
		} else {
			
			//	Find the interesting element.
			
			Element* iter = this->head;
			Count    skip = pos-1;
			
			while (iter && skip) {
				iter = iter->next;
				skip--;
			}
			
			//	Remove at current position.
			
			(elem->next ? elem->next->prev : this->tail) = elem->prev;
			(elem->prev ? elem->prev->next : this->head) = elem->next;
			
			//	Insert after "iter".
			
			elem->prev = iter;
			elem->next = iter->next;
			
			(iter->next ? iter->next->prev : this->tail) = elem;
			iter->next = elem;
		}
		
		return TRUE;
	}

	//	NIY : updating the iterator...
	
	return FALSE;
}





/*
 *	Return the state of the list. If it has been sorted once
 *	and if the insertions have also taken into account the
 *	sorting, then the list still is sorted.
 */

Bool
OPList::IsSorted () const
{
	return this->is_sorted;
}


/*
 *	Define insertion mode. Default is automatic sorting while
 *	inserting elements.
 */

Bool
OPList::SetInsertMode (Bool sort)
{
	this->sort_on_insert = sort;
	return TRUE;
}

Bool
OPList::GetInsertMode () const
{
	return this->sort_on_insert;
}


Bool
OPList::PushHead (OPObjectIn value)
{
	OP_METHOD (PushHead, 1);
	OP_ARG (1, OPObjectIn, value);
	
	Element* elem  = NEW_FROM_ZONE (this->ReturnMemZone ()) Element;
	
	//	Insert at the head of the list...
	
	elem->next = this->head;
	elem->prev = 0;
	
	elem->data = value;
	
	(elem->prev ? elem->prev->next : this->head) = elem;
	(elem->next ? elem->next->prev : this->tail) = elem;
	
	this->is_sorted = FALSE;
	this->num_element++;
	
	if (this->cache_elem) {
		this->cache_index++;
	}
	
	if (this->iterators.IsValid ()) {
		Count index = 0;
		OPListIteratorRef iter;
		
		while (this->iterators->Get (index++, iter)) {
			if (iter.IsValid ()) {
				iter->NotifyInsert (elem);
			}
		}
	}
	
	return TRUE;
}

Bool
OPList::PushTail (OPObjectIn value)
{
	OP_METHOD (PushHead, 1);
	OP_ARG (1, OPObjectIn, value);
	
	Element* elem  = NEW_FROM_ZONE (this->ReturnMemZone ()) Element;
	
	//	Insert at the head of the list...
	
	elem->next = 0;
	elem->prev = this->tail;
	
	elem->data = value;
	
	(elem->prev ? elem->prev->next : this->head) = elem;
	(elem->next ? elem->next->prev : this->tail) = elem;
	
	this->is_sorted = FALSE;
	this->num_element++;
	
	if (this->iterators.IsValid ()) {
		Count index = 0;
		OPListIteratorRef iter;
		
		while (this->iterators->Get (index++, iter)) {
			if (iter.IsValid ()) {
				iter->NotifyInsert (elem);
			}
		}
	}
	
	return TRUE;
}

Bool
OPList::PopHead (OPObjectOut value)
{
	OP_METHOD (PopHead, 1);
	OP_ARG (1, OPObjectOut, value);
	
	Element* elem = this->head;
	
	if (elem) {
		
		if (this->iterators.IsValid ()) {
			Count index = 0;
			OPListIteratorRef iter;
			
			while (this->iterators->Get (index++, iter)) {
				if (iter.IsValid ()) {
					iter->NotifyRemove (elem);
				}
			}
		}
		
		if (this->cache_elem == elem) {
			this->cache_elem  = 0;
			this->cache_index = 0;
		}
		if (this->cache_elem) {
			this->cache_index--;
		}
		
		(elem->next ? elem->next->prev : this->tail) = elem->prev;
		(elem->prev ? elem->prev->next : this->head) = elem->next;
		this->num_element--;
		
		value = elem->data;
		elem->data = 0;
		
		DELETE (elem);
		
		return TRUE;
	}
	
	return FALSE;
}

Bool
OPList::PopTail (OPObjectOut value)
{
	OP_METHOD (PopTail, 1);
	OP_ARG (1, OPObjectOut, value);
	
	Element* elem = this->tail;
	
	if (elem) {
		
		if (this->iterators.IsValid ()) {
			Count index = 0;
			OPListIteratorRef iter;
			
			while (this->iterators->Get (index++, iter)) {
				if (iter.IsValid ()) {
					iter->NotifyRemove (elem);
				}
			}
		}
		
		(elem->next ? elem->next->prev : this->tail) = elem->prev;
		(elem->prev ? elem->prev->next : this->head) = elem->next;
		this->num_element--;
		
		value = elem->data;
		elem->data = 0;
		
		DELETE (elem);
		
		return TRUE;
	}
	
	return FALSE;
}
