/*
 *	scrgrid.cxx
 *
 *	The Scroll Grid implements a support for a large cell grid
 *	including the addition of scrollers.
 *
 *	(C) Copyright 1997-1998, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#include <OPaC/cellgrid.h>
#include <OPaC/grail.h>
#include <OPaC/lookman.h>
#include <OPaC/new.h>
#include <OPaC/pubmacro.h>
#include <OPaC/scrgrid.h>
#include <OPaC/scroller.h>
#include <OPaC/scrview.h>
#include <OPaC/targetmacro.h>

/****************************************************************/

OP_IMPLEMENT_CLASS (OPScrollGrid);

/****************************************************************/

OP_TARGET_ACTOR (OPScrollGrid, GridChangedSize);

OP_PUBLISH_MODIFIER (OPScrollGrid, "CellGrid", GetCellGrid, SetCellGrid, "Cell grid", OPObject, OPCellGrid);
OP_PUBLISH_MODIFUNC (OPScrollGrid, "PermanentHScroller", ReturnPermanentHScroller, SetPermanentHScroller, "Permanent horizontal scroller", Bool, Bool);
OP_PUBLISH_MODIFUNC (OPScrollGrid, "PermanentVScroller", ReturnPermanentVScroller, SetPermanentVScroller, "Permanent vertical scroller", Bool, Bool);

/****************************************************************/

OPScrollGrid::OPScrollGrid ()
{
	OP_METHOD (OPScrollGrid, 0);
	
	this->shape &= ~ OP_CTRL_RELIEF;
	this->shape |= OP_CTRL_NOFRAME;
	
	this->is_valid = FALSE;
	
	this->permanent_h_scroller = FALSE;
	this->permanent_v_scroller = FALSE;
}

OPScrollGrid::~OPScrollGrid ()
{
}

void
OPScrollGrid::Delete ()
{
	OP_METHOD (Delete, 0);
	
	this->scroll_view = 0;
	this->h_scroller  = 0;
	this->v_scroller  = 0;
	this->grid        = 0;
	this->is_valid    = FALSE;
	
	super::Delete ();
}

void
OPScrollGrid::StoreInstance (OPStorageIn storage)
{
	OP_METHOD (StoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	super::StoreInstance (storage);
	
	storage << this->grid;
	storage << this->scroll_view;
	storage << this->h_scroller << this->v_scroller;
	storage << this->permanent_h_scroller;
	storage << this->permanent_v_scroller;
	storage << this->is_valid;
	storage << OPStorageSync::mark;
}

void
OPScrollGrid::RestoreInstance (OPStorageIn storage)
{
	OP_METHOD (RestoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	super::RestoreInstance (storage);
	
	storage >> this->grid;
	storage >> this->scroll_view;
	storage >> this->h_scroller >> this->v_scroller;
	storage >> this->permanent_h_scroller;
	storage >> this->permanent_v_scroller;
	storage >> this->is_valid;
	storage >> OPStorageSync::mark;
}



/*
 *	Initialise the scroll view, the scroll bars and
 *	the cell grid so that everything works.
 */

void
OPScrollGrid::Initialise (OPCellGridIn grid, OPRectIn box)
{
	OP_METHOD (Initialise, 2);
	OP_ARG (1, OPCellGridIn, grid);
	OP_ARG (2, OPRectIn, box);
	
	OP_ASSERT (this->grid.IsValid () == FALSE);
	
	//	Allocate the resources needed by the scroll grid :
	//
	//	- two scrollers (horizontal and vertical)
	//	- a scroll view to hold the grid
	
	this->h_scroller  = OP_NEW (OPScroller);
	this->v_scroller  = OP_NEW (OPScroller);
	this->scroll_view = OP_NEW (OPScrollView);
	this->grid        = grid;
	
	OPCoord breadth  = this->h_scroller->ReturnBreadth ();
	OPCoord total_dx = 0;
	OPCoord total_dy = 0;
	
	this->grid->GetSize (total_dx, total_dy);
	
	//	Set the size of the scroll grid to the box specified by the caller.
	//	Compute the approximative remaining size and position the elements.
	
	this->SetBox (box);
	
	OPCoord x  = box.X1 ()+1;
	OPCoord y  = box.Y1 ()+1;
	OPCoord dx = box.DX ()-2 - breadth;
	OPCoord dy = box.DY ()-2 - breadth;
	
	this->scroll_view->SetParent (this);
	this->scroll_view->SetOrigin (x, y);
	this->scroll_view->SetSize (dx, dy);
	
	this->grid->SetParent (this->scroll_view);
	this->grid->SetOrigin (x, y);
	this->grid->SetDrawFrameHigh (FALSE);
	this->grid->SetDrawFrameLow (FALSE);
	this->grid->AttachObserver (this, "SizeChanged", "GridChangedSize");
	
	this->h_scroller->SetParent (this);
	this->h_scroller->Initialise (x, y+dy, dx, TRUE);
	
	this->v_scroller->SetParent (this);
	this->v_scroller->Initialise (x+dx, y, dy, FALSE);
	
	this->scroll_view->SetHScroller (this->h_scroller);
	this->scroll_view->SetVScroller (this->v_scroller);
	this->scroll_view->SetTopLeft (0, 0);
	this->scroll_view->SetTotalSize (total_dx, total_dy);
	
	this->is_valid = TRUE;
	
	//	Compute the real size of the elements, decide whether the scrollers
	//	will be shown or not, etc.
	
	this->GridChangedSize ();
}


/*
 *	Define whether the horizontal and vertical scroller are
 *	permanent or not (in which case, they won't be displayed
 *	when they are not needed).
 */

Bool
OPScrollGrid::SetPermanentVScroller (Bool yes)
{
	OP_METHOD (SetPermanentVScroller, 1);
	OP_ARG (1, Bool, yes);
	
	this->permanent_v_scroller = yes;
	
	return TRUE;
}

Bool
OPScrollGrid::SetPermanentHScroller (Bool yes)
{
	OP_METHOD (SetPermanentHScroller, 1);
	OP_ARG (1, Bool, yes);
	
	this->permanent_h_scroller = yes;
	
	return TRUE;
}

Bool
OPScrollGrid::ReturnPermanentVScroller () const
{
	OP_METHOD (ReturnPermanentVScroller, 0);
	return this->permanent_v_scroller;
}

Bool
OPScrollGrid::ReturnPermanentHScroller () const
{
	OP_METHOD (ReturnPermanentVScroller, 0);
	return this->permanent_h_scroller;
}



/*
 *	Change the cell grid represented in the scroll view. This
 *	is only possible if the scroll grid has been initialised
 *	once before.
 */

Bool
OPScrollGrid::SetCellGrid (OPCellGridIn grid)
{
	OP_METHOD (SetCellGrid, 1);
	OP_ARG (1, OPCellGridIn, grid);
	
	OP_ASSERT (this->scroll_view.IsValid ());
	OP_ASSERT (this->grid.IsValid ());
	OP_ASSERT (grid.IsValid ());
	
	OPCoord x = this->box.X1 () + 1;
	OPCoord y = this->box.Y1 () + 1;
	
	if (this->grid.IsValid ()) {
		this->grid->GetOrigin (x, y);
		this->grid->SetParent (OPViewNull);
	}
	
	this->grid = grid;
	
	this->grid->SetParent (this->scroll_view);
	this->grid->SetOrigin (x, y);
	this->grid->SetDrawFrameHigh (FALSE);
	this->grid->SetDrawFrameLow (FALSE);
	
	//	Update the dimensions and the scroll positions...
	
	this->GridChangedSize ();
	
	return TRUE;
}


/*
 *	Return the current cell grid represented in this scroll
 *	view.
 */

Bool
OPScrollGrid::GetCellGrid (OPCellGridOut grid) const
{
	OP_METHOD (GetCellGrid, 1);
	OP_ARG (1, OPCellGridOut, grid);
	
	grid = this->grid;
	
	return TRUE;
}


/*
 *	When the scroll grid size changes, we will need to recompute
 *	the size of each element (scroll view, scrollers).
 */

void
OPScrollGrid::ChangeGeometry (OPRectIn box)
{
	OP_METHOD (ChangeGeometry, 1);
	OP_ARG (1, OPRectIn, box);
	
	super::ChangeGeometry (box);
	
	if (this->is_valid) {
		this->GridChangedSize ();
	}
}


/*
 *	The size of either the cell grid or the scroll grid changed.
 *	Find out whether the scrollers are really needed. If not, just
 *	remove them.
 */

void
OPScrollGrid::GridChangedSize ()
{
	OP_METHOD (GridChangedSize, 0);
	
	if (this->is_valid == FALSE) {
		return;
	}
	
	OPCoord total_dx;
	OPCoord total_dy;
	OPCoord breadth = this->h_scroller->ReturnBreadth ();
	OPCoord x       = box.X1 ()+1;
	OPCoord y       = box.Y1 ()+1;
	OPCoord box_dx  = this->box.DX ()-2;
	OPCoord box_dy  = this->box.DY ()-2;
	
	this->grid->GetSize (total_dx, total_dy);
	this->scroll_view->SetTotalSize (total_dx, total_dy);
	
	//	Decide whether the scrollers will be needed or not. If a scroller can
	//	be hidden, there will be more room in the other direction to display
	//	the scroll view.
	
	Bool has_x_scroller = TRUE;
	Bool has_y_scroller = TRUE;
	
	if (this->permanent_h_scroller) {
		box_dy -= breadth;
	}
	if (this->permanent_v_scroller) {
		box_dx -= breadth;
	}
	
	if ( (total_dx <= box_dx)
	  && (total_dy <= box_dy) ) {
		has_x_scroller = this->permanent_h_scroller;
		has_y_scroller = this->permanent_v_scroller;
	} else {
		if (this->permanent_h_scroller == FALSE) box_dy -= breadth;
		if (this->permanent_v_scroller == FALSE) box_dx -= breadth;
		
		if (total_dx <= box_dx) {
			if (this->permanent_h_scroller == FALSE) {
				has_x_scroller = FALSE;
				box_dy += breadth;
			}
		}
		if (total_dy <= box_dy) {
			if (this->permanent_v_scroller == FALSE) {
				has_y_scroller = FALSE;
				box_dx += breadth;
			}
		}
	}
	
	//	Show or hide the scrollers, depending on the geometry.
	
	this->h_scroller->SetParent (has_x_scroller ? this : 0);
	this->v_scroller->SetParent (has_y_scroller ? this : 0);
	
	//	Set the size/position depending on the scroller's presence.
	
	this->scroll_view->SetOrigin (x, y);
	this->scroll_view->SetSize (box_dx, box_dy);
	
	if (has_x_scroller) {
		this->h_scroller->SetOrigin (x, y+box_dy);
		this->h_scroller->SetSize (box_dx, breadth);
	} else {
		this->h_scroller->SetVisibleValue (0);
		this->scroll_view->Scroll (this->h_scroller, 0);
	}
	
	if (has_y_scroller) {
		this->v_scroller->SetOrigin (x+box_dx, y);
		this->v_scroller->SetSize (breadth, box_dy);
	} else {
		this->v_scroller->SetVisibleValue (0);
		this->scroll_view->Scroll (this->v_scroller, 0);
	}
}


/*
 *	
 */

void
OPScrollGrid::Draw (OPGrailPortIn port, OPRectIn rect, OPDrawOp what)
{
	OP_METHOD (Draw, 3);
	OP_ARG (1, OPGrailPortIn, port);
	OP_ARG (2, OPRectIn, rect);
	OP_ARG (3, OPDrawOp, what);
	
	const OPAdornerRef& adorner = port->ReturnAdorner ();
	const OPLookManRef& lookman = OPLookMan::Current ();
	
	OPCoord margin = (this->shape & OP_CTRL_NOFRAME) ? 0 : lookman->ReturnWidgetFrameSize ();
	OPCoord effect = (this->shape & OP_CTRL_RELIEF) ? lookman->Return3DEffectWidth () : 0;
	
	if (what & OP_DRAW_BACKGROUND) {
		adorner->FillRect (this->box, OP_COLOR_BACKGROUND);
	}
	if (what & OP_DRAW_OBJECT) {
		adorner->DrawRect (this->box, OP_COLOR_BLACK);
	}
	
	super::Draw (port, rect, what);
}
