/*
 *	scroller.cxx
 *
 *	Scroller class definition. The scroller may be represented
 *	vertically or horizontally by a bar, a knob and buttons.
 *
 *	(C) Copyright 1994-1998, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#include <OPaC/scroller.h>
#include <OPaC/icon.h>
#include <OPaC/lookman.h>
#include <OPaC/grail.h>
#include <OPaC/event.h>
#include <OPaC/window.h>
#include <OPaC/except.h>
#include <OPaC/new.h>
#include <OPaC/targetmacro.h>
#include <OPaC/pubmacro.h>
#include <OPaC/notmacro.h>

/****************************************************************/

OP_IMPLEMENTATION (scroller.h);
OP_IMPLEMENT_CLASS (OPScroller);

/****************************************************************/

#define	SCROLL_DECR		((Card32)(1))
#define	SCROLL_INCR		((Card32)(2))

/****************************************************************/

/*
 *	Export the miscellanous targets supported by the OPaC
 *	Scroller class :
 */

OP_TARGET_OBSERVER (OPScroller, OnClick);
OP_TARGET_OBSERVER (OPScroller, OnContClick);

/****************************************************************/

OP_PUBLISH_MODIFUNC (OPScroller, "VisibleValue", ReturnVisibleValue, SetVisibleValue, "Visible value", Real, Real);
OP_PUBLISH_MODIFUNC (OPScroller, "VisibleRatio", ReturnVisibleRatio, SetVisibleRatio, "Visible ratio", Real, Real);
OP_PUBLISH_MODIFUNC (OPScroller, "Increment", ReturnIncrement, SetIncrement, "Increment Step", Real, Real);

/****************************************************************/

OP_NOTIFY_EVENT (OPScroller, Scroll);
OP_NOTIFY_EVENT (OPScroller, ScrollBegin);
OP_NOTIFY_EVENT (OPScroller, ScrollEnd);

OP_NOTIFY_CHANGE (OPScroller, VisibleValue, Scroll);
OP_NOTIFY_CHANGE (OPScroller, VisibleValue, Changed);
OP_NOTIFY_CHANGE (OPScroller, VisibleRatio, Changed);

/****************************************************************/

/*
 *	Global Scroller class variables :
 */

Bool    OPScroller::look_inited    = FALSE;
OPCoord OPScroller::look_breadth   = 0;
OPCoord OPScroller::look_frame     = 0;
OPCoord OPScroller::look_3D_effect = 0;
Count   OPScroller::look_repeat    = 0;
Count   OPScroller::look_start     = 0;

/****************************************************************/

/*
 *	The scroller automatically sets up two scroll buttons
 *	which will be used to scroll up/down automatically.
 */

OPScroller::OPScroller ()
{
	OP_METHOD (OPScroller, 0);
	
	//	Override state defined by the OPView class :
	
	this->state |= OP_VIEW_IS_ENABLED;
	this->state |= OP_VIEW_IS_ALIVE;
	this->state &= ~ OP_VIEW_IS_DISABLED;
	this->state &= ~ OP_VIEW_OK_MAIN;
	this->shape |= OP_CTRL_NOTACTENTER;
	
	//	Override some definitions made by the OPControl class :
	
	this->do_hilite_on_action       = TRUE;
	this->do_press_on_action        = FALSE;
	this->do_cont_reports_on_action = FALSE;
	
	//	Initialise the scroller specific fields :
	
	this->value = 0.0;
	this->ratio = 1.0;
	this->increment = 0.1;
	
	this->knob_min   = 0;
	this->knob_max   = 0;
	
	this->knob_pos   = 0;
	this->knob_size  = 0;
	this->knob_click = 0;
	
	this->tick_count = 0;
	
	this->is_horiz   = FALSE;
    this->is_knob_ok = TRUE;
    this->is_ok      = FALSE;
    
	this->init_arrow_pos = OPLookMan::Current ()->ReturnScrollArrowPos ();
	this->real_arrow_pos = OPLookMan::Current ()->ReturnScrollArrowPos ();
	
	//	There are no arrows available. They will be added by the explicit
	//	initialization method :
	
	this->arrow_incr = 0;
	this->arrow_decr = 0;
	
	if (this->look_inited == FALSE) {
		
		//	Initialise the look :
	
		const OPLookManRef& look = OPLookMan::Current ();
		
		OPCoord dx = 0;
		OPCoord dy = 0;
		
		look->GetIconSize (OP_ICON_LEFT_NORMAL, dx, dy);
		
		this->look_breadth   = (dx > dy) ? dx : dy;
		this->look_frame     = look->ReturnWidgetFrameSize ();
		this->look_3D_effect = look->Return3DEffectWidth ();
		this->look_repeat    = look->ReturnAutoRepeatTime ();
		this->look_start     = look->ReturnAutoStartTime ();
		this->look_inited    = TRUE;
	}
}



/*
 *	Clean up the scroller : delete both buttons. There is
 *	no need to disconnect the buttons; this will be done
 *	automatically.
 */

OPScroller::~OPScroller ()
{
	OP_METHOD (~OPScroller, 0);
	
	this->arrow_incr = 0;
	this->arrow_decr = 0;
}



/*
 *	Store the scroller...
 */

void
OPScroller::StoreInstance (OPStorageIn storage)
{
	OP_METHOD (StoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	super::StoreInstance (storage);
	
	storage << this->value << this->ratio << this->increment;
	storage << this->arrow_incr << this->arrow_decr << this->init_arrow_pos;
	storage << this->is_horiz;
	storage << OPStorageSync::mark;
}



/*
 *	Restore the scroller...
 */

void
OPScroller::RestoreInstance (OPStorageIn storage)
{
	OP_METHOD (RestoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	super::RestoreInstance (storage);

	storage >> this->value >> this->ratio >> this->increment;
	storage >> this->arrow_incr >> this->arrow_decr >> this->init_arrow_pos;
	storage >> this->is_horiz;
	storage >> OPStorageSync::mark;
}



/*
 *	Prepare the scroller for destruction. Give up both
 *	arrow buttons.
 */

void
OPScroller::Delete ()
{
	OP_METHOD (Delete, 0);
	
	if (this->arrow_incr.IsValid ()) {
		this->arrow_incr->Delete ();
		this->arrow_incr = 0;
	}
	
	if (this->arrow_decr.IsValid ()) {
		this->arrow_decr->Delete ();
		this->arrow_decr = 0;
	}
	
	super::Delete ();
}



/*
 *	After the scroller has been restored (eg from archive)
 *	it will be awoken by one of the following methods :
 */

void
OPScroller::Awake ()
{
	OP_METHOD (Awake, 0);
	
	super::Awake ();
	this->CheckSpaceForParts ();
}

void
OPScroller::AwakeFromApp (OPAppIn app)
{
	OP_METHOD (AwakeFromApp, 1);
	OP_ARG (1, OPAppIn, app);
	
	OPControl::AwakeFromApp (app);
	this->CheckSpaceForParts ();
}

void
OPScroller::AwakeFromObject (OPObjectIn object)
{
	OP_METHOD (AwakeFromObject, 1);
	OP_ARG (1, OPObjectIn, object);
	
	OPControl::AwakeFromObject (object);
	this->CheckSpaceForParts ();
}


/*
 *	Update the current knob size (depending on the visibility
 *	ratio).
 */

void
OPScroller::UpdateKnobSize ()
{
	OP_METHOD (UpdateKnobSize, 0);
	
	OPCoord len  = this->knob_max - this->knob_min + 1;
	OPCoord knob = this->look_breadth - this->look_frame * 2;
	
	this->knob_size = (OPCoord)(this->ratio * len);
	
	if (this->knob_size < knob) {
		
		//	Don't allow the knob to shrink too much, or it won't be easy to
		//	grab for the user.
		
		this->knob_size = knob;
	}
}


/*
 *	Compute the new button and knob positions/sizes after
 *	a geometry modification.
 */

void
OPScroller::CheckSpaceForParts ()
{
	OP_METHOD (CheckSpaceForParts, 0);
	
	//	Find out the possible sizes for the knob. If there is not enough room
	//	for a knob and two arrow buttons, then we suppress the arrows.
	
	this->knob_min = (this->is_horiz ? this->box.X1 () : this->box.Y1 ());
	this->knob_max = (this->is_horiz ? this->box.X2 () : this->box.Y2 ());
	
	OPCoord mean_size = this->knob_max - this->knob_min - 2 * this->look_frame;
	OPCoord mini_size = this->look_breadth * 3;
	OPCoord maxi_size = this->look_breadth * 5;
	
	this->real_arrow_pos = (mean_size < mini_size) ? OP_SCROLL_ARROW_NONE
												   : this->init_arrow_pos;
	
	if ( (this->arrow_incr.IsValid ())
	  && (this->arrow_decr.IsValid ()) ) {
		
		if ( (this->real_arrow_pos == OP_SCROLL_ARROW_MINMAX)
		  && (mean_size < maxi_size) ) {
			
			//	Not enough room for two pairs of arrows (top and bottom). Just
			//	remove half of the arrow buttons :
			
			this->real_arrow_pos = OP_SCROLL_ARROW_NORMAL;
		}
		
		//	Update the knob extrema according to the arrow buttons.
		
		switch (this->real_arrow_pos) {
			
			case OP_SCROLL_ARROW_NONE:
				this->knob_max -= this->look_frame;
				this->knob_min += this->look_frame;
				break;
			
			case OP_SCROLL_ARROW_MAX:
				this->knob_max -= this->look_breadth * 2;
				this->knob_min += this->look_frame;
				break;
			
			case OP_SCROLL_ARROW_MIN:
				this->knob_max -= this->look_frame;
				this->knob_min += this->look_breadth * 2;
				break;
			
			case OP_SCROLL_ARROW_NORMAL:
				this->knob_min += this->look_breadth;
				this->knob_max -= this->look_breadth;
				break;
			
			default:
				OP_ABORT ("Unsupported arrow position.");
		}
	}
	
	//	Compute the knob position, using the values we just obtained.
	
	this->UpdateKnobSize ();
	
	OPCoord len  = this->knob_max - this->knob_min + 1;
	this->knob_pos = (OPCoord)(this->value * (len - this->knob_size));
	
	
	//	Change the position of the arrow buttons so that they will coincide
	//	with the scroller's geometry :
	
	if (this->is_horiz) {
		
		OPCoord x1 = this->box.X1 ();
		OPCoord x2 = this->box.X2 () - this->look_breadth + 1;
		OPCoord y  = this->box.Y1 ();
		
		switch (this->real_arrow_pos) {
			
			case OP_SCROLL_ARROW_MAX:
				x1 = x2 - this->look_breadth;
				break;
			
			case OP_SCROLL_ARROW_MIN:
				x2 = x1 + this->look_breadth;
				break;
			
			default:
				break;
		}
		
		if (this->arrow_decr.IsValid ()) {
			this->arrow_decr->SetOrigin (x1, y);
		}
		
		if (this->arrow_incr.IsValid ()) {
			this->arrow_incr->SetOrigin (x2, y);
		}
		
	} else {
		
		OPCoord x  = this->box.X1 ();
		OPCoord y1 = this->box.Y1 ();
		OPCoord y2 = this->box.Y2 () - this->look_breadth + 1;
		
		switch (this->real_arrow_pos) {
			
			case OP_SCROLL_ARROW_MAX:
				y1 = y2 - this->look_breadth;
				break;
			
			case OP_SCROLL_ARROW_MIN:
				y2 = y1 + this->look_breadth;
				break;
			
			default:
				break;
		}
		
		if (this->arrow_decr.IsValid ()) {
			this->arrow_decr->SetOrigin (x, y1);
		}
		if (this->arrow_incr.IsValid ()) {
			this->arrow_incr->SetOrigin (x, y2);
		}
	}
}



/*
 *	Draw the scroller's knob. I am still not sure if this
 *	solution is good or if the knob should be implemented
 *	as a separate object.
 */

void
OPScroller::DrawKnob (OPGrailPortIn port)
{
	OP_METHOD (DrawKnob, 1);
	OP_ARG (1, OPGrailPortIn, port);
	
	OP_ASSERT (this->is_knob_ok);
	
	//	The adorner will be used to do the drawing. This provides some easy way
	//	to change the overall look of OPaC applications.
	
	const OPAdornerRef& adorner = port->ReturnAdorner ();
	
	//	Compute the bounding box where the knob may be located. This depends on
	//	the scroller's bounding box and on the buttons which have been used.
	
	OPCoord x1 = (this->is_horiz) ? (this->knob_min) : (this->box.X1 () + this->look_frame);
	OPCoord y1 = (this->is_horiz) ? (this->box.Y1 () + this->look_frame) : (this->knob_min);
	OPCoord x2 = (this->is_horiz) ? (this->knob_max) : (this->box.X2 () - this->look_frame);
	OPCoord y2 = (this->is_horiz) ? (this->box.Y2 () - this->look_frame) : (this->knob_max);
	
	OPRect inside (x1, y1, x2, y2);
	
	//	Fill the background in order to be sure we erased the previously drawn
	//	knob, which may have moved.
	
	adorner->FillRect (inside, OP_COLOR_WIDGET_BACK);
	
	//	Now find out where the knob has to be drawn. The knob may be pressed
	//	and/or hilited...
	
	(this->is_horiz ? x1 : y1) += this->knob_pos;
	
	OPCoord w  = this->look_breadth - 2 * this->look_frame;
	OPCoord dx = this->is_horiz ? this->knob_size : w;
	OPCoord dy = this->is_horiz ? w : this->knob_size;
	
	inside.SetOrigin (x1, y1);
	inside.SetSize (dx, dy);
	
	this->is_being_pressed ? adorner->DrawRectLow (inside)
						   : adorner->DrawRectHigh (inside);
	
	if (this->is_being_hilited) {
		adorner->FillRect (inside, OP_COLOR_HILITE_CLICK, this->look_3D_effect);
	} else {
		adorner->FillRect (inside, OP_COLOR_WIDGET_FORE, this->look_3D_effect);
	}
}


/*
 *	Draw the scroller's body.
 */

void
OPScroller::DrawBody (OPGrailPortIn port)
{
	OP_METHOD (DrawBody, 1);
	OP_ARG (1, OPGrailPortIn, port);
	
	const OPAdornerRef& adorner = port->ReturnAdorner ();
	adorner->DrawRect (this->box, OP_COLOR_BLACK);
}



/*
 *	Initialise an arrow button.
 */

void
OPScroller::InitArrow (OPIconInOut arrow, Card32 id)
{
	OP_METHOD (InitArrow, 1);
	OP_ARG (1, OPIconInOut, arrow);
	
	if (arrow.IsValid () == FALSE) {
		
		//	Allocate the arrow and connect them to the scroller :
		
		arrow = OP_NEW_FROM_ZONE (this->ReturnMemZone (), OPIcon);
		
		arrow->SetParent (this);
		arrow->SetReportMode (TRUE);
		arrow->SetFlagOkMain (FALSE);
		arrow->ChangeState (OP_VIEW_OK_UI_EDIT, FALSE);
		arrow->ChangeState (OP_VIEW_OK_UI_MOVE, FALSE);
		
		//	Connect the buttons to the scroller so that the messaging will
		//	work automatically :
		
		OPScrollerRef self = this;
		
		arrow->AttachObserver (self, "OnClick", id);
		arrow->AttachObserver (self, "OnContClick", id);
	}
}



/*
 *	Initialise the icons after having allocated the arrows.
 */

void
OPScroller::InitIcons (Bool is_horiz)
{
	OP_METHOD (InitIcons, 1);
	OP_ARG (1, Bool, is_horiz);
	
	if (is_horiz) {
		
		this->arrow_decr->Initialise (OP_ICON_LEFT_NORMAL, OP_ICON_LEFT_PRESS,
									  OP_ICON_LEFT_HILITE, OP_ICON_LEFT_DIM);
		
		this->arrow_incr->Initialise (OP_ICON_RIGHT_NORMAL, OP_ICON_RIGHT_PRESS,
									  OP_ICON_RIGHT_HILITE, OP_ICON_RIGHT_DIM);
		
	} else {
		
		this->arrow_decr->Initialise (OP_ICON_UP_NORMAL, OP_ICON_UP_PRESS,
									  OP_ICON_UP_HILITE, OP_ICON_UP_DIM);
		
		this->arrow_incr->Initialise (OP_ICON_DOWN_NORMAL, OP_ICON_DOWN_PRESS,
									  OP_ICON_DOWN_HILITE, OP_ICON_DOWN_DIM);
	}
	
	this->arrow_decr->SetDefaultSize ();
	this->arrow_incr->SetDefaultSize ();
}


/*
 *	Initialise the scroller: setup geometry and insert the
 *	arrow buttons.
 */

void
OPScroller::Initialise (OPCoord x, OPCoord y, OPCoord size, Bool is_horiz)
{
	OP_METHOD (Initialise, 4);
	OP_ARG (1, OPCoord, x);
	OP_ARG (2, OPCoord, y);
	OP_ARG (3, OPCoord, size);
	OP_ARG (4, Bool, is_horiz);
	
	this->InitArrow (this->arrow_incr, SCROLL_INCR);
	this->InitArrow (this->arrow_decr, SCROLL_DECR);
	
	//	Ok, connections done. Now initialise the arrows.
	
	OPCoord dx = is_horiz ? size : this->look_breadth;
	OPCoord dy = is_horiz ? this->look_breadth : size;
	
	this->InitIcons (is_horiz);

	this->is_horiz = is_horiz;
	
	this->SetOrigin (x, y);
	this->SetSize (dx, dy);
		
	this->CheckSpaceForParts ();
	this->is_ok = TRUE;
}



/*
 *	Set the arrow button's position.
 */

void
OPScroller::SetArrowPos (OPScrollArrowPos pos)
{
	OP_METHOD (SetArrowPos, 1);
	OP_ARG (1, OPScrollArrowPos, pos);
	
	this->init_arrow_pos = pos;
	this->CheckSpaceForParts ();
}


Bool
OPScroller::SetIncrement (Real increment)
{
	OP_METHOD (SetIncrement, 1);
	OP_ARG (1, Real, increment);
	this->increment = increment;
	return TRUE;
}


/*
 *	Set value which should be displayed by the scroller.
 *	It is used to determine the knob's position.
 */

Bool
OPScroller::SetVisibleValue (Real value)
{
	OP_METHOD (SetVisibleValue, 1);
	OP_ARG (1, Real, value);
	
	if (value < 0.0) value = 0.0;
	if (value > 1.0) value = 1.0;
	
	if (this->value == value) {
		return TRUE;
	}
	
	this->value = value;
	this->CheckSpaceForParts ();
	this->ReportChange ();
	
	return TRUE;
}



/*
 *	Set ratio of knob size versus scroller lenth. A 1.0
 *	ratio corresponds to a whole length knob.
 */

Bool
OPScroller::SetVisibleRatio (Real ratio)
{
	OP_METHOD (SetVisibleRatio, 1);
	OP_ARG (1, Real, ratio);
	
	if (ratio < 0.0) ratio = 0.0;
	if (ratio > 1.0) ratio = 1.0;
	
	if (this->ratio == ratio) {
		return TRUE;
	}
	
	this->ratio = ratio;
	this->CheckSpaceForParts ();
	this->ReportChange ();
	
	return TRUE;
}



/*
 *	Return the current visible value. See SetVisibleValue.
 */

Real
OPScroller::ReturnVisibleValue () const
{
	OP_METHOD (ReturnVisibleValue, 0);
	return this->value;
}



/*
 *	Return the current visible ratio. See SetVisibleRatio.
 */

Real
OPScroller::ReturnVisibleRatio () const
{
	OP_METHOD (ReturnVisibleRatio, 0);
	return this->ratio;
}


Real
OPScroller::ReturnIncrement () const
{
	OP_METHOD (ReturnIncrement, 0);
	return this->increment;
}

/*
 *	Return the breadth (either width or height, depending on the
 *	orientation of the scroller).
 */

OPCoord
OPScroller::ReturnBreadth () const
{
	OP_METHOD (ReturnBreadth, 0);
	return this->look_breadth;
}


/****************************************************************/

/*
 *	This observer gets called whenever one of the buttons associated
 *	to the scroller is pressed by the user. The "id" determines which
 *	button has been pressed.
 */

void
OPScroller::OnClick (OPObjectIn source, Card32 id)
{
	OP_METHOD (OnClick, 2);
	OP_ARG (1, OPObjectIn, source);
	OP_ARG (2, Card32, id);
	
	this->tick_count = 0;
	this->Scroll (id);
}

/*
 *	The continuous click observer is notified at regular intervals
 *	once a button has been pressed, until the user releases it.
 */

void
OPScroller::OnContClick (OPObjectIn source, Card32 id)
{
	OP_METHOD (OnContClick, 2);
	OP_ARG (1, OPObjectIn, source);
	OP_ARG (2, Card32, id);
	
	this->tick_count += this->look_repeat;
	
	if (this->tick_count > this->look_start) {
		this->Scroll (id);
	}
}

/****************************************************************/

/*
 *	Scroll the knob, since the user pressed one of the
 *	arrow buttons.
 */

void
OPScroller::Scroll (Card32 id)
{
	OP_METHOD (Scroll, 1);
	OP_ARG (1, Card32, id);
	
	Real value = this->value;
	
	if (id == SCROLL_DECR) {
		
		value -= this->increment;
		if (value < 0.0) value = 0.0;
		
	} else if (id == SCROLL_INCR) {
		
		value += this->increment;
		if (value > 1.0) value = 1.0;
	}
	
	if (value != this->value) {
		
		//	There really was some scrolling (the position of the knob changed)
		//	and we will have to update the scroller's aspect and tell every
		//	observer that we "scrolled" !
		
		this->SetVisibleValue (value);
		this->Report (notify_Scroll, OP_REPORT_NEW);
		this->Display ();
		this->ReturnAdorner ()->Flush (OP_FLUSH_SYNC);
	}
}

/****************************************************************/

/*
 *	Resizing a scroller might change its shape, since the
 *	arrows may get suppressed.
 */

void
OPScroller::ChangeGeometry (OPRectIn box)
{
	OP_METHOD (ChangeGeometry, 1);
	OP_ARG (1, OPRectIn, box);
	
	super::ChangeGeometry (box);
	this->CheckSpaceForParts ();
}



/*
 *	Return the scroller's minimal size. One dimension is
 *	always constrained, the other is free.
 */

void
OPScroller::GetMinSize (OPCoordOut dx, OPCoordOut dy) const
{
	OP_METHOD (GetMinSize, 2);
	OP_ARG (1, OPCoordOut, dx);
	OP_ARG (2, OPCoordOut, dy);
	
	dx = this->look_breadth * (this->is_horiz ? 3 : 1);
	dy = this->look_breadth * (this->is_horiz ? 1 : 3);
}



/*
 *	Return the scroller's maximum size. Return an arbitrary
 *	huge value.
 */

void
OPScroller::GetMaxSize (OPCoordOut dx, OPCoordOut dy) const
{
	OP_METHOD (GetMaxSize, 2);
	OP_ARG (1, OPCoordOut, dx);
	OP_ARG (2, OPCoordOut, dy);
	
	super::GetMaxSize (dx, dy);
	(this->is_horiz ? dy : dx) = this->look_breadth;
}



/*
 *	Draw the scroller : just draw the knob and the frame.
 *	The arrows will be drawn by the super-class !
 */

void
OPScroller::Draw (OPGrailPortIn port, OPRectIn rect, OPDrawOp what)
{
	OP_METHOD (Draw, 3);
	OP_ARG (1, OPGrailPortIn, port);
	OP_ARG (2, OPRectIn, rect);
	OP_ARG (3, OPDrawOp, what);
	
//	if (what & OP_DRAW_BACKGROUND) {
//		adorner->FillRect (this->box, OP_COLOR_WIDGET_BACK, 1);
//		what &= ~OP_DRAW_BACKGROUND;
//	}
	
	super::Draw (port, rect, what);
	
	if (what & OP_DRAW_OBJECT) {
		this->DrawBody (port);
		this->DrawKnob (port);
	}
}



/*
 *	When a mouse button gets pressed, there will be quite
 *	some work to do.
 */

void
OPScroller::ButtonPress (OPEventIn event)
{
	OP_METHOD (ButtonPress, 1);
	OP_ARG (1, OPEventIn, event);
	
	if ( (this->state & OP_VIEW_IS_ENABLED)
	  && (this->is_being_clicked == FALSE)
	  && (event->ReturnChange () == OP_EV_BUTTON_LEFT) ) {
		
		OPCoord len = this->knob_max - this->knob_min + 1;
		OPCoord dec = this->knob_pos;
		OPCoord inc = len - this->knob_pos - this->knob_size;
		OPCoord pos = this->is_horiz ? event->ReturnX () : event->ReturnY ();
		OPCoord rel = pos - this->knob_min;
		
		OPScrollerRef self = this;
		
		this->window->SetDraggingView (self);
		this->ActionOn ();
		this->Report (notify_ScrollBegin, OP_REPORT_NEW);
		this->Display ();
		
		if ( (rel < dec)
		  || (rel > (len - inc)) ) {
			
			//	The user did not click into the knob...
			
			//	NIY :	political choice : what to do when user clicks into the
			//			scroller's body... (jump/fast scroll/scroll/beep)
			
			this->knob_click = (OPRelCoord)(this->knob_size / 2);
			this->DragNotify (event);
			
		} else {
			
			this->knob_click = (OPRelCoord)(rel - this->knob_pos);
		}
	}
}



/*
 *	Drag the knob while the user moves around the mouse
 *	with the left button pressed.
 */

void
OPScroller::DragNotify (OPEventIn event)
{
	OP_METHOD (DragNotify, 1);
	OP_ARG (1, OPEventIn, event);
	
	if (this->is_being_clicked) {
		
		OPCoord len = this->knob_max - this->knob_min + 1 - this->knob_size;
		OPCoord pos = this->is_horiz == TRUE ? event->ReturnX () : event->ReturnY ();
		OPCoord rel = pos - this->knob_min;
		
		pos = rel - this->knob_click;
		
		if (pos < 0) pos = 0;
		if (pos > len) pos = len;
		
		if ((OPCoord)(pos) != this->knob_pos) {
			
			Real value = (Real)(pos) / (Real)(len);
			
			this->knob_pos = (OPCoord)(pos);
			
			this->SetVisibleValue (value);
			this->Report (notify_Scroll, OP_REPORT_NEW);
			this->Display ();
			this->ReturnAdorner ()->Flush (OP_FLUSH_SYNC);
		}
	}
}



/*
 *	The user released a mouse button. If it is the left
 *	one, finish the current knob dragging operation.
 */

void
OPScroller::ButtonRelease (OPEventIn event)
{
	OP_METHOD (ButtonRelease, 1);
	OP_ARG (1, OPEventIn, event);
	
	if ( (this->state & OP_VIEW_IS_ENABLED)
	  && (this->is_being_clicked == TRUE)
	  && (event->ReturnChange () == OP_EV_BUTTON_LEFT) ) {
		
		OPScrollerRef nobody = 0;
		
		this->window->SetDraggingView (nobody);
		this->Report (notify_ScrollEnd, OP_REPORT_NEW);
		this->ActionOff ();
	}
}



/*
 *	The Look Manager notifies the view instances whenever
 *	the user changed its look ! The view need not redraw
 *	itself now, since the window will do it automatically.
 */

void
OPScroller::Broadcast (OPAtomIn name, OPObjectIn param)
{
	OP_METHOD (Broadcast, 2);
	OP_ARG (1, OPAtomIn, name);
	OP_ARG (1, OPObjecIn, param);
	
	super::Broadcast (name, param);
	
	if (name == this->atom_look_man_changed) {
	
		OPLookManRef look = param;
		
		OPCoord dx = 0;
		OPCoord dy = 0;
		
		look->GetIconSize (OP_ICON_LEFT_NORMAL, dx, dy);
		
		OP_ASSERT (dx == dy);
		
		this->look_breadth   = dx;
		this->look_frame     = look->ReturnWidgetFrameSize ();
		this->look_3D_effect = look->Return3DEffectWidth ();
		this->look_repeat    = look->ReturnAutoRepeatTime ();
		this->look_start     = look->ReturnAutoStartTime ();
		this->look_inited    = TRUE;
		
		if (this->is_ok) {
			
			//	Re-initialise the whole scroller using the new geometry, but without
			//	moving its origin.
			
			OPCoord x    = this->box.X1 ();
			OPCoord y    = this->box.Y1 ();
			OPCoord size = this->is_horiz ? this->box.DX () : this->box.DY ();
			
			this->Initialise (x, y, size, this->is_horiz);
		}
	}
}
