/*
 *	set.cxx
 *
 *	Base set classes : abstract set and sorted dictionary. These are NOT optimized
 *	at all and will have to be implemented in a better way as soon as possible :-(.
 *
 *	(C) Copyright 1994-1997, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#include <OPaC/array.h>
#include <OPaC/container.h>
#include <OPaC/atom.h>

/****************************************************************/

OP_ABSTRACT_CLASS (OPSet);

/****************************************************************/

#define EXTRACT_ANY_KEY			0
#define EXTRACT_ANY_DATA		1
#define EXTRACT_DATA_FROM_KEY	2
#define EXTRACT_KEY_FROM_DATA	3


	
OPSet::OPSet ()
{
	OP_METHOD (OPSet, 0);
	
	this->num_element = 0;
	this->extr_mode   = EXTRACT_ANY_KEY;
	this->extr_key    = OPAtomNull;
	this->extr_data   = OPAtomNull;
}




OPSet::~OPSet ()
{
	OP_METHOD (~OPSet, 0);
	OP_ASSERT (this->num_element == 0);
}




void
OPSet::StoreInstance (OPStorageIn storage)
{
	OP_METHOD (StoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	super::StoreInstance (storage);
	
	storage << this->num_element
	        << this->extr_mode
	        << this->extr_key
			<< this->extr_data;
	storage << OPStorageSync::mark;
}




void
OPSet::RestoreInstance (OPStorageIn storage)
{
	OP_METHOD (RestoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	super::RestoreInstance (storage);
	
	storage >> this->num_element
	        >> this->extr_mode
	        >> this->extr_key
			>> this->extr_data;
	storage >> OPStorageSync::mark;
}



/*
 *	Prepare the set for destruction.
 */

void
OPSet::Delete ()
{
	OP_METHOD (Delete, 0);
	
	this->extr_mode = EXTRACT_ANY_KEY;
	this->extr_key  = OPAtomNull;
	this->extr_data = OPAtomNull;
	
	this->RemoveAll ();
	
	super::Delete ();
}



/*
 *	Specify a generic extraction. We will have to interpret
 *	the command in order to be able to understand what kind
 *	of extraction we will have to perform...
 *
 *	These are the accepted extractions :
 *
 *	 - "ExtractValue"
 *   - "ExtractKey"
 *	 - "ExtractValueUsingKey" : OPObjectIn key
 *	 - "ExtractKeyUsingValue" : OPObjectIn value
 *
 *	If the extraction category does not match a listed one,
 *	then FALSE is returned.
 */

Bool
OPSet::SpecifyExtraction (OPArrayIn x)
{
	OP_METHOD (SpecifyExtraction, 1);
	OP_ARG (1, OPArrayIn, x);
	
	static OPAtomRef x_value       = "ExtractValue";
	static OPAtomRef x_key         = "ExtractKey";
	static OPAtomRef x_value_using = "ExtractValueUsingKey";
	static OPAtomRef x_key_using   = "ExtractKeyUsingValue";

	Count size = x->TotalCount ();
	
	if ( (size == 0)
	  || (x[0].IsValid () == FALSE)
	  || (x[0]->IsKindOf (OP_RUNTIME (OPAtom)) == FALSE) ) {
		return FALSE;
	}
	
	OPAtomRef how = x[0];

	if ( (how  == x_value)
	  && (size == 1) ) {
		
		//	Extract any value : specify no key for extraction.
		
		OPObjectRef none;
		this->SpecifyExtractKey (none);
		return TRUE;
	}

	if ( (how  == x_key)
	  && (size == 1) ) {
		
		//	Extract any key : specify no value for extraction.
		
		OPObjectRef none;
		this->SpecifyExtractValue (none);
		return TRUE;
	}

	if ( (how  == x_value_using)
	  && (size == 2) ) {
		
		//	Extract value using the specified key.
		
		this->SpecifyExtractKey (x[1]);
		return TRUE;
	}

	if ( (how  == x_key_using)
	  && (size == 2) ) {
		
		//	Extract key using the specified value.
		
		this->SpecifyExtractValue (x[1]);
		return TRUE;
	}

	return FALSE;
}



/*
 *	Specify the extraction key. If no key is specified, then
 *	every value will be extracted...
 */

void
OPSet::SpecifyExtractKey (OPObjectIn key)
{
	OP_METHOD (SpecifyExtractKey, 1);
	OP_ARG (1, OPObjectIn, key);

	this->extr_mode = (key.IsValid ()) ? EXTRACT_DATA_FROM_KEY : EXTRACT_ANY_DATA;
	this->extr_key  = key;
	this->extr_data = OPAtomNull;
}



/*
 *	Specify the extraction value. If no value is specified,
 *	then every key will be extracted...
 */

void
OPSet::SpecifyExtractValue (OPObjectIn data)
{
	OP_METHOD (SpecifyExtractValue, 1);
	OP_ARG (1, OPObjectIn, data);

	this->extr_mode = (data.IsValid ()) ? EXTRACT_KEY_FROM_DATA : EXTRACT_ANY_KEY;
	this->extr_key  = OPAtomNull;
	this->extr_data = data;
}



/*
 *	Return the total number of elements stored in the set.
 */

Count
OPSet::TotalCount () const
{
	OP_METHOD (TotalCount, 0);
	
	return this->num_element;
}



/*
 *	Generic "Get" which allows to extract either the keys
 *	or the data, with or without a key/value.
 */

Bool
OPSet::Get (Count pos, OPObjectOut object)
{
	OP_METHOD (Get, 2);
	OP_ARG (1, Count, pos);
	OP_ARG (2, OPObjectOut, object);
	
	Bool        ok   = FALSE;
	OPObjectRef key  = this->extr_key;
	OPObjectRef data = this->extr_data;
	OPObjectRef none;
	
	switch (this->extr_mode) {
		
		case EXTRACT_ANY_KEY:
			ok     = this->Get (pos, key, data);
			object = ok ? key : none;
			break;

		case EXTRACT_KEY_FROM_DATA:
			ok     = this->GetFromValue (data, key, pos);
			object = ok ? key : none;
			break;

		case EXTRACT_ANY_DATA:
			ok     = this->Get (pos, key, data);
			object = ok ? data : none;
			break;

		case EXTRACT_DATA_FROM_KEY:
			ok     = this->GetFromKey (key, data, pos);
			object = ok ? data : none;
			break;

	}
	
	return ok;
}

