/*
 *	time.cxx
 *
 *	Time representation and handling class.
 *
 *	(C) Copyright 1998, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#include <OPaC/time.h>

/****************************************************************/

OP_IMPLEMENT_CLASS (OPTime);

/****************************************************************/

static inline Int32
floor (Int32 a, Int32 b)               /*  assumes b positive  */
{
	return (a >= 0) ? (a/b) : (a%b == 0 ? 0 : -1) - labs (a)/b;
}

static const int
g_month_length[2][14] =
{
	{0, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31, 0 },
	{0, 31, 29, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31, 0 }
};

static inline Bool
is_leap_year (Card32 year, Bool is_gregorian = TRUE)
{
	return (year % 4) || (is_gregorian && ((year % 100) == 0) && (year % 400)) ? FALSE : TRUE;
}

/****************************************************************/

/*
 *	Convert a standard date into a Gregorian day number. The
 *	zero is set to 15 October 1582, date when the calendar
 *	came into effect.
 *
 *	The input date can be either Gregorian or Julian.
 *
 *	Algorithm taken from Peter J.G. Meyer, DDJ, March 1993.
 */

static Bool
date_to_gdn (int day, int month, int year, Int32& gdn, Bool is_gregorian = TRUE)
{
	Bool leap = is_leap_year (year, is_gregorian);
	
	if ( (month < 1)
	  || (month > 12)
	  || (day < 1)
	  || (day > g_month_length[leap][month]) ) {
		return FALSE;
	}
	
	/*  calculate number of days before/after October 15, 1582 (Gregorian) */
    
	gdn = (year-1)*365 + floor (year-1, 4);
    
	if (is_gregorian) {
		gdn += floor (year-1, 400) - floor (year-1, 100);
	}
	while (--month) {
		gdn += g_month_length[leap][month];
	}
	
	gdn += day - 577736 - (is_gregorian ? 0 : 2);
    
    return TRUE;
}

/*
 *	Convert the Gregorian day number into a standard date. The
 *	output date can be either Gregorian or Julian.
 */

static void
gdn_to_date (int& day, int& month, int& year, Int32 gdn, Bool is_gregorian = TRUE)
{
	gdn += 577735 + (is_gregorian ? 0 : 2);
	
	Int32 y400 = 146100 - (is_gregorian ? 0 : 3);
	Int32 y100 =  36525 - (is_gregorian ? 0 : 1);
	Int32 y4   =   1461;
	
	Bool exception = FALSE;
	
	year =  400 * floor (gdn, y400);
	gdn -= y400 * floor (gdn, y400);
	
	if (gdn > 0) {
		
		year +=  100 * floor (gdn, y100);
		gdn  -= y100 * floor (gdn, y100);
		
		exception = (gdn == 0) && (is_gregorian);
		
		if (gdn > 0) {
			
			year +=  4 * floor (gdn, y4);
			gdn  -= y4 * floor (gdn, y4);
			
			if (gdn > 0) {
				
				int i = 0;
				
				while ( (gdn > 365) && (i++ < 3) ) {
					year++;
					gdn -= 365;
                }
            }
        }
    }
	
	if (exception) {
		gdn = 366;			/*  occurs once every hundred years with Gregorian calendar  */
	} else {
		year++;
		gdn++;
    }
	
	Bool leap = is_leap_year (year, is_gregorian);
	
	month = 1;
	while ( (month < 13)
		 && (gdn > g_month_length[leap][month]) ) {
		gdn -= g_month_length[leap][month++];
		if (month == 13) {
			month = 1;
			year++;
		}
	}
}

/*
 *	Returns day of week for given Gregorian day number.
 *	0=Sunday, 6=Saturday
 */

static int
day_of_week (Int32 gdn)
{
	return ((int)(((gdn%7)+12)%7));
}



OPTime::OPTime ()
{
	OP_METHOD (OPTime, 0);
	
	this->day  = 0;
	this->time = 0;
}

OPTime::~OPTime ()
{
	OP_METHOD (~OPTime, 0);
}


void
OPTime::StoreInstance (OPStorageIn storage)
{
	OP_METHOD (StoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	super::StoreInstance (storage);
	
	storage << this->day << this->time;
	storage << OPStorageSync::mark;
}

void
OPTime::RestoreInstance (OPStorageIn storage)
{
	OP_METHOD (RestoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	super::RestoreInstance (storage);
	
	storage >> this->day >> this->time;
	storage >> OPStorageSync::mark;
}


Bool
OPTime::SetDate (Count day, Count month, Count year)
{
	OP_METHOD (SetDate, 3);
	OP_ARG (1, Count, day);
	OP_ARG (2, Count, month);
	OP_ARG (3, Count, year);
	
	Int32 gdn;
	
	if (date_to_gdn (day, month, year, gdn)) {
		this->day = gdn;
		return TRUE;
	}
	
	return FALSE;
}

Bool
OPTime::GetDate (CountOut day, CountOut month, CountOut year) const
{
	OP_METHOD (GetDate, 3);
	OP_ARG (1, CountOut, day);
	OP_ARG (2, CountOut, month);
	OP_ARG (3, CountOut, year);
	
	int d, m, y;
	gdn_to_date (d, m, y, this->day);
	
	day   = d;
	month = m;
	year  = y;
	
	return TRUE;
}

Bool
OPTime::SetTime (Count h, Count m, Count s, Count ms)
{
	OP_METHOD (SetTime, 4);
	OP_ARG (1, Count, h);
	OP_ARG (2, Count, m);
	OP_ARG (3, Count, s);
	OP_ARG (4, Count, ms);
	
	if ( (h < 24)
	  && (m < 60)
	  && (s < 60)
	  && (ms < 1000) ) {
		this->time = ms + 1000*(s+60*(m+60*h));
		return TRUE;
	}
	
	return FALSE;
}

Bool
OPTime::GetTime (CountOut h, CountOut m, CountOut s) const
{
	OP_METHOD (GetTime, 3);
	OP_ARG (1, CountOut, h);
	OP_ARG (2, CountOut, m);
	OP_ARG (3, CountOut, s);
	
	Count ms;
	return this->GetTime (h, m, s, ms);
}


Bool
OPTime::GetTime (CountOut h, CountOut m, CountOut s, CountOut ms) const
{
	OP_METHOD (GetTime, 4);
	OP_ARG (1, CountOut, h);
	OP_ARG (2, CountOut, m);
	OP_ARG (3, CountOut, s);
	OP_ARG (4, CountOut, ms);
	
	Card32 time = this->time;
	
	ms    = time % 1000;
	time /= 1000;
	s     = time % 60;
	time /= 60;
	m     = time % 60;
	time /= 60;
	h     = time;
	
	return TRUE;
}

Bool
OPTime::SetAbsTime (Int64 time)
{
	OP_METHOD (SetAbsTime, 1);
	OP_ARG (1, Int64, time);
	
	Int64 ms  = time % (1000*60*60*24);
	Int64 day = time / (1000*60*60*24);
	
	this->time = (Card32)(ms);
	this->day  = (Card32)(day);
	
	return TRUE;
}


Int64
OPTime::ReturnAbsTime () const
{
	OP_METHOD (ReturnAbsTime, 0);
	
	Int64 ms  = this->time;
	Int64 day = this->day;
	
	return ms + day * (1000*60*60*24);
}


Count
OPTime::ReturnDayOfWeek () const
{
	OP_METHOD (ReturnDayOfWeek, 0);
	return day_of_week (this->day);
}

/****************************************************************/

Int64
OPTime::DateTimeToAbsTime (Count day, Count month, Count year, Count h, Count m, Count s, Count ms)
{
	Int32 gdn;
	
	if (date_to_gdn (day, month, year, gdn)) {
		Int64 x_ms  = ms+1000*(s+60*(m+60*h));
		Int64 x_day = gdn * 1000*60*60*24;
		return x_ms + x_day;
	}
	
	return 0;
}

void
OPTime::AbsTimeToDateTime (Int64 a, CountOut day, CountOut month, CountOut year, CountOut h, CountOut m, CountOut s)
{
	Count ms;
	AbsTimeToDateTime (a, day, month, year, h, m, s, ms);
}

void
OPTime::AbsTimeToDateTime (Int64 a, CountOut day, CountOut month, CountOut year, CountOut hour, CountOut min, CountOut sec, CountOut ms)
{
	Int64 x_ms  = a % (1000*60*60*24);
	Int64 x_day = a / (1000*60*60*24);
	
	Card32 time = (Card32)(x_ms);
	Card32 date = (Card32)(x_day);
	
	ms    = time % 1000;
	time /= 1000;
	sec   = time % 60;
	time /= 60;
	min   = time % 60;
	time /= 60;
	hour  = time;
	
	int d, m, y;
	gdn_to_date (d, m, y, date);
	
	day   = d;
	month = m;
	year  = y;
}

Count
OPTime::AbsTimeToDayOfWeek (Int64 a)
{
	return day_of_week ((Int32)(a / (1000*60*60*24)));
}

/****************************************************************/
