/*
 *	gfxhelpers.h
 *
 *	Graphic library : general helper functions/macros.
 *
 *	(C) Copyright 1998, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#ifndef	_GFX_GFXHELPERS_H_
#define	_GFX_GFXHELPERS_H_

/****************************************************************/

#define	GFX_ALIGN_BITS	(4)
#define	GFX_ALIGN_SIZE	(1<<GFX_ALIGN_BITS)
#define	GFX_ALIGN_MASK	(0xFFFFFFFF<<GFX_ALIGN_BITS)

/****************************************************************/

extern const Card32 g_map_5_to_8[];
extern const Card32 g_map_6_to_8[];

/****************************************************************/

/*
 *	Align the size to the default alignment proposed for the current
 *	architecture.
 */

static inline Card32
q_gfx_align (Card32 s)
{
	return (s + GFX_ALIGN_SIZE-1) & GFX_ALIGN_MASK;
}

/*
 *	Clip the input rectangle so that it fits in the box.
 */

static inline void
q_gfx_cliprect (Int32& x, Int32& y, Card32& dx, Card32& dy, Card32 box_dx, Card32 box_dy)
{
	if (x >= (Int32)box_dx) {
		dx = 0;
	} else if (x+dx >= box_dx) {
		dx = box_dx - x;
	}
	
	if (y >= (Int32)box_dy) {
		dy = 0;
	} else if (y+dy >= box_dy) {
		dy = box_dy - y;
	}
}

static inline void
q_gfx_cliprect (Int32& x,  Int32& y,  Card32& dx, Card32& dy,
				Card32 x1, Card32 y1, Card32 x2, Card32 y2)
{
	if ( (x >= (Int32)x2)
	  || (x+(Int32)dx <= (Int32)x1) ) {
		dx = 0;
	} else {
		if (x+(Int32)dx >= (Int32)x2) {
			dx = x2 - x;
		}
		if (x < (Int32)x1) {
			dx += x - x1;
			x  -= x - x1;
		}
	}
	
	if ( (y >= (Int32)y2)
	  || (y+(Int32)dy <= (Int32)y1) ) {
		dy = 0;
	} else {
		if (y+(Int32)dy >= (Int32)y2) {
			dy = y2 - y;
		}
		if (y < (Int32)y1) {
			dy += y - y1;
			y  -= y - y1;
		}
	}
}

/*
 *	Test if the input rectangle has any pixels in it.
 */

static inline Bool
q_gfx_testrect (Card32 dx, Card32 dy)
{
	return (dx && dy);
}

/*
 *	Replicate the pixel value in order to fill 32 bits.
 */

static inline void
q_gfx_replicatepixel (Card32& pixel, Card32 num)
{
	switch (num) {
		case 1: pixel = (pixel << 8) | (pixel & 0xFF);
		case 2: pixel = (pixel << 16) | (pixel & 0xFFFF); break;
	}
}

static inline Card8
q_gfx_rgb_to_gray (Card32 pixel)
{
	Card32 r = ((pixel >> 16) & 0xFF);
	Card32 g = ((pixel >>  8) & 0xFF);
	Card32 b = ((pixel >>  0) & 0xFF);
	return (Card8)((r*307+g*614+b*103) >> 10);
}

static inline Card16
q_gfx_rgb_to_565 (Card32 pixel)
{
	Card32 r = (pixel >> 16);
	Card32 g = (pixel >>  8);
	Card32 b = (pixel & 0xFF);
	return (Card16)(((r & 0xF8) << 8) | ((g & 0xFC) << 3) | (b >> 3));
}

static inline Card16
q_gfx_rgb_to_565 (Card32 r, Card32 g, Card32 b)
{
	return (Card16)(((r & 0xF8) << 8) | ((g & 0xFC) << 3) | (b >> 3));
}

static inline Card8
q_gfx_blend_rgb_to_gray (Card8 d_pix, Card32 s_pix)
{
	Card32 alpha = s_pix >> 24;
	Card32 gray  = q_gfx_rgb_to_gray (s_pix);
	
	if (alpha == 255) {
		return d_pix;
	}
	if (alpha == 0) {
		return (Card8)(gray);
	}
	if (alpha > 127) alpha++;
	
	Card32 ap = alpha;
	Card32 an = 256 - ap;
	
	return (Card8)( (gray * an + d_pix * ap) >> 8);
}

static inline Card16
q_gfx_blend_rgb_to_565 (Card16 d_pix, Card32 s_pix)
{
	Card32 alpha = s_pix >> 24;
	
	if (alpha == 255) {
		return d_pix;
	}
	if (alpha == 0) {
		return q_gfx_rgb_to_565 (s_pix);
	}
	if (alpha > 127) alpha++;
	
	Card32 ap = alpha;
	Card32 an = 256 - ap;
	Card32 sr = ((s_pix >> 16) & 0xFF);
	Card32 sg = ((s_pix >>  8) & 0xFF);
	Card32 sb = ((s_pix >>  0) & 0xFF);
	Card32 dr = ((g_map_5_to_8[(d_pix >> 11) & 0x1F]) * ap + sr * an) >> 8;
	Card32 dg = ((g_map_6_to_8[(d_pix >>  5) & 0x3F]) * ap + sg * an) >> 8;
	Card32 db = ((g_map_5_to_8[(d_pix >>  0) & 0x1F]) * ap + sb * an) >> 8;
	
	return q_gfx_rgb_to_565 (dr, dg, db);
}

static inline Card32
q_gfx_blend_rgb_to_rgb (Card32 d_pix, Card32 s_pix)
{
	Card32 alpha = s_pix >> 24;
	
	if (alpha == 255) {
		return d_pix;
	}
	if (alpha == 0) {
		d_pix &= 0xFF000000;
		d_pix |= s_pix & 0x00FFFFFF;
		return d_pix;
	}
	if (alpha > 127) alpha++;
	
	Card32 ap = alpha;
	Card32 an = 256 - ap;
	Card32 r  = ((d_pix & 0x00FF0000) * ap + (s_pix & 0x00FF0000) * an) & 0xFF000000;
	Card32 g  = ((d_pix & 0x0000FF00) * ap + (s_pix & 0x0000FF00) * an) & 0x00FF0000;
	Card32 b  = ((d_pix & 0x000000FF) * ap + (s_pix & 0x000000FF) * an) & 0x0000FF00;
	d_pix &= 0xFF000000;
	d_pix |= (r | g | b) >> 8;
	return d_pix;
}


static inline Card32
q_gfx_gray_to_rgb (Card32 pixel, Card32 alpha)
{
	return (((((alpha << 8) | pixel) << 8) | pixel) << 8) | pixel;
}

static inline Card32
q_gfx_565_to_rgb (Card32 pixel, Card32 alpha)
{
	return (alpha << 24) | (g_map_5_to_8[(pixel >> 11) & 0x1F])
						 | (g_map_6_to_8[(pixel >>  5) & 0x3F])
						 | (g_map_5_to_8[(pixel >>  0) & 0x1F]);
}

/****************************************************************/

#endif /* _GFX_GFXHELPERS_H_ */
